import numpy as np
import lmfit as lf
import glob
import scipy.signal as sps
from scipy import interpolate as interp
import pandas as pd

class Resonator(object):
    r"""Fit an S21 measurement of a hanger (or notch) type resonator.

    Parameters
    ----------
    name : string
        The resonator name. Does not have to be unique, but each physical
        resonator in the experiment should have a unique name to avoid
        confusion when using some of the other tools in scraps.

    temp : float
        The temperature in (K) that the S21 measurement was taken at.

    pwr : float
        The power (in dBm) at which the resonator was measured.

    freq : array-like[nDataPoints]
        The frequency points at which the S21 scan was measured.

    I : array-like[nDataPoints]
        The in-phase (or real part) of the complex S21 measurement. Units
        are typically volts, and I should be specified in linear units (as
        opposed to dB).

    Q : array-like[nDataPoints]
        The out-of-phase (or imaginary part) of the complex S21 measurement.
        Units are typically volts, and I should be specified in linear units
        (as opposed to dB).

    sigmaI : array-like[nDataPoints] (optional)
        An array of uncertaintly values for each data point in `I`. Default
        is ``None``.

    sigmaQ : array-like[nDataPoints] (optional)
        An array of uncertaintly values for each data point in `Q`. Default
        is ``None``.

    The following attributes are automatically calculated and added during
    initialization.

    Attributes
    ----------
    name : string
        The resonator name passed at initialization.

    temp : float
        The temperature passed at initialization.

    pwr : float
        The power passed at initialization.

    freq : array-like[nDataPoints]
        The frequency points passed at initialization.

    I : array-like[nDataPoints]
        The I data points passed at initialization.

    Q : array-like[nDataPoints]
        The Q data points passed at initialization.

    sigmaI : array-like[nDataPoints]
        The sigmaI values passed at initialization.

    sigmaQ : array-like[nDataPoints]
        The sigmaQ values passed at initialization.

    S21 : array-like[nDataPoints]
        The complex transmission ``S21 = I + 1j*Q``.

    phase : array-like[nDataPoints]
        The raw phase ``phase = np.arctan2(Q, I)``.

    uphase : array-like[nDataPoints]
        The unwrapped phase is equivalent to the phase, but with jumps of 2 Pi
        removed.

    mag : array-like[nDataPoints]
        The magnitude ``mag = np.abs(S21)`` or, equivalently ``mag =
        np.sqrt(I**2 + Q**2)``.

    hasFit : bool
        Indicates whether or not ``Resonator.do_lmfit`` method has been called.

    lmfit_result : ``lmfit.Result`` object
        The result object created by ``lmfit`` containing all the fit
        information. Some of the fit information is futher extracted for
        convenience in the following Attributes. For an exhaustive list of the
        attributes of lmfit_result see the docs for ``lmfit``. The most useful
        attribute of this object is ``lmfit_result.params``, which contains the
        best-fit parameter values.

    residualI : array-like[nDataPoints]
        The residual of the fit model against the `I` data, wieghted by the
        uncertainties.

    residualQ : array-like[nDataPoints]
        The residual of the fit model against the `Q` data, wieghted by the
        uncertainties.

    resultI : array-like[nDataPoints]
        The best ``lmfit`` fit result to the fit model for `I`.

    resultQ : array-like[nDataPoints]
        The best ``lmfit`` fit result to the fit model for `Q`.

    resultMag : array-like[nDataPoints]
        ``resultMag = np.abs(resultI + 1j*resultQ)``

    resultPhase : array-like[nDataPoints]
        ``resultPhase = np.arctan2(resultQ/resultI)``

    emcee_result : ``lmfit.Result`` object
        This object is nearly identical to the `lmfit_result` object, but also
        contains the maximum-liklihood values for the *varying* parameters of
        the fit model as well as the `chains` returned by ``emcee``. The most
        important attribute is probably ``emcee_result.flatchain``, which can be
        passed directly to ``pygtc`` or ``corner`` to make a really nice
        GTC/Triangle/Corner plot. For an exhaustive list of the attributes of
        emcee_result see the docs for ``lmfit``, specifically the section
        involving the ``lmfit`` implementation of ``emcee``.

    mle_vals : list of float
        The maximum-liklihood estimate values of the *varying* parameter in the
        fit model as calculated by ``emcee``. Unpacked here for convenience from
        ``emcee_result.params``.

    mle_labels: list of string
        The parameter names of the values in `mle_vals`. Provided here for easy
        passing to  ``pygtc`` or ``corner``.

    magBaseLine : array-like[nDataPoints]
        The best initial guess of the baseline of the magnitude. Calculated by
        fitting a quadratic polynomial to the beginning and end of the magnitdue
        vs frequency curve.

    phaseBaseLine: array-like[nDataPoints]
        The best initial guess of the baseline of the phase. Calculated by
        fitting a line to the beginning and end of the phase vs frequency curve.
        This is equivalent to calculating the electrical delay in the
        measurement lines.

    params : ``lmfit.Parameters`` object
        The initial parameter guesses for fitting the `S21` data. See ``lmfit``
        documentation for a complete overview. To get the parameter names, call
        ``params.keys()``. Default is ``None``. Initialize params by calling
        ``Resonator.load_params``. Delete params with
        ``Resonator.torch_params``.

    """


    #Do some initialization
    def __init__(self, name, temp, pwr, freq, I, Q, sigmaI = None, sigmaQ = None,
                 I0 = None, Q0 = None,caldata = None):
        r"""Initializes a resonator object by calculating magnitude, phase, and
        a bunch of fit parameters for a hanger (or notch) type S21 measurement.

        """
        self.name = name
        self.temp = temp
        self.pwr = pwr
        self.freq = np.asarray(freq)
        self.I = np.asarray(I)
        self.Q = np.asarray(Q)
        self.sigmaI = np.asarray(sigmaI) if sigmaI is not None else None
        self.sigmaQ = np.asarray(sigmaQ) if sigmaQ is not None else None
        
        self.I0 = I0  if I0 is not None else np.zeros(np.size(self.I))
        self.Q0 = Q0  if Q0 is not None else np.zeros(np.size(self.Q))
        
        self.caldata = caldata
        
        self.S21 = I + 1j*Q
        self.phase = np.arctan2(Q,I) #use arctan2 because it is quadrant-aware
        self.uphase = np.unwrap(self.phase) #Unwrap the 2pi phase jumps
        self.mag = np.abs(self.S21) #Units are volts.
        self.logmag = 20*np.log10(self.mag) #Units are dB (20 because V->Pwr)

        #Find the frequency at magnitude minimum (this can, and should, be
        #overwritten by a custom params function)
        self.fmin = self.freq[np.argmin(self.mag)]

        #Whether or not params has been initialized
        self.params = None
        self.hasParams = False

        #These won't exist until the lmfit method is called
        self.lmfit_result = None

        #These are scheduled for deprecation. They will eventually live in the lmfit_result dictionary
        self.hasFit = False
        self.hasnoisedata = False
        self.residualI = None
        self.residualQ = None
        self.resultI = None
        self.resultQ = None
        self.resultMag = None
        self.resultPhase = None
        
        #The noise parameters
        
        self.noisefreq = []
        self.noiseIs = []
        self.noiseQs = []
        self.noiseIsoffres = []
        self.noiseQsoffres = []
        # self.noisevoffsets = []
        # self.noisevgains = []
        self.noiseUnscaled = []
        self.noiseScaled = []
        self.noiseOffresUnscaled = []
        self.noiseOffresScaled = []
        
        
        
        #The pulse parameters
        self.pulseWidths = []
        self.pulseTimes = []
        self.pulseAmps = []
        self.pulsePhases = []
        self.pulseIs = []
        self.pulseQs = []
        self.pulseERIs = []
        self.pulseERQs = []
        # self.pulsevoffsets = []
        # self.pulsevgains = []
        self.pulseUnscaled = []
        self.pulseScaled = []
        self.DCOffsetSweepData = []
        self.IQSweepData = []
        

        #These won't exist until the emcee method is called
        self.emcee_result = None

        #These are scheduled for deprecation. They will eventually live in the lmfit_result dictionary
        self.hasChain = False
        self.mle_vals = None
        self.mle_labels = None

    def to_disk(self):
        """To be implemented: dumps resonator to disk as various file types. Default will be netcdf4"""
        pass

    def from_disk(self):
        """To be implemented: load resonator object from disk."""
        pass

    def to_json(self):
        """To be implemented: serialize resonator as a JSON string"""
        pass

    def from_json(self):
        """To be implemented: create rsonator from JSON string"""
        pass

    #TODO: Implement the following for handling pickling:

    #def __getstate__(self):
    #   pass

    #def __setstate__(self):
    #   pass


    def load_params(self, paramsFn, **kwargs):
        """Load up a lmfit Parameters object for a custom fit function.

        Parameters
        ----------
        paramsFn : method
            The paramsFn method should return a ``lmfit.Paramters`` object. This
            object will be passed to the fit method when ``do_lmfit`` or
            ``do_emcee`` is
            called.

        kwargs : dict
            A dictionary of keyword arguments to pass to paramsFn.

        """
        params = paramsFn(self, **kwargs)
        self.params = params
        self.hasParams = True
    
    def load_noise(self,noisedata,noisefreq,onres = True):
        
        if onres == True:
            
            self.onresnoisefreq = noisefreq;
            self.onresnoisedata = noisedata
            self.hasonresnoisedata = True
            
        if onres == False:
            self.offresnoisefreq = noisefreq
            self.offresnoisedata = noisedata
            self.hasoffresnoisedata = True
            
    def cal_noise_spectrum_absolute(self,noisedata,noisefreqmeas,onres = True,nperseg = 125000,fs = 5e6,
                           combinespectrum = False):
        
        I = np.real(noisedata)
        Q = np.imag(noisedata)
        
        freqnoise, Sxx = sps.welch(I - np.mean(I),fs = fs,window = 'hann',nperseg = nperseg)
        freqnoise, Syy = sps.welch(Q - np.mean(Q),fs = fs,window = 'hann',nperseg = nperseg)
        
        
        if onres == True:
            self.freqOnResNoise = freqnoise;
            self.SxxOnResAb = Sxx;
            self.SyyOnResAb = Syy;
            
        else:
            self.freqOffResNoise = freqnoise;
            self.SxxOffResAb = Sxx;
            self.SyyOffResAb = Syy;
        
    def cal_noise_spectrum(self,noisedata,noisefreqmeas,onres = True,nperseg = 125000,fs = 5e6,
                           combinespectrum = False,method = 'welch'):
        
        #combinespectrum: set the spectrum to be 
        #   1Hz   for f < 50Hz, 
        #   10Hz  for 50Hz<f<1kHz 
        #   100Hz for 1kHz<f<1MHz
        #   1kHz  for f>1MHz
        
        # if self.hasnoisedata == True:
        indx = np.argmin(np.abs(self.freq-noisefreqmeas))
        
        #the delay of the transmission 
        baselinenoise = self.resultIBaseline[indx] + 1j*self.resultQBaseline[indx]
        
        #calibrate the noise data
        noisedata = noisedata/baselinenoise
        
        # S21_noise = self.S21[indx];
        S21_noise = self.resultINorm[indx] + 1j*self.resultQNorm[indx]
        
        xc = self.x0
        yc = self.y0
        r0 = self.r0
    
        #assume that the resonance is a quarter wavelength resonator
    
        angle = np.angle(S21_noise - (xc + 1j*yc))
    
    
        if angle <0 :
            rotate_angle =   - (np.pi + np.angle(S21_noise - (xc + 1j*yc))); #the rotation angle
        else:
            rotate_angle =  np.pi - np.angle(S21_noise - (xc + 1j*yc));
    
        #the following action has already move the center to the origin and rotated resonance circle
        
        if onres == True:
            self.ZNoiseOnRes = noisedata[0:10000]
            
        else:
            self.ZNoiseOffRes = noisedata[0:10000]
    
    
        # rotate noise on the x axis
        I_rotated = (np.real(noisedata) - xc)*np.cos(rotate_angle) - (np.imag(noisedata) - yc)*np.sin(rotate_angle)
        Q_rotated = (np.real(noisedata) - xc)*np.sin(rotate_angle) + (np.imag(noisedata) - yc)*np.cos(rotate_angle)
    
        del noisedata
        #normalize the noise with radius 
        I_rotated = I_rotated/r0; 
        Q_rotated = Q_rotated/r0;
        
        # I_rotated = RemoveGlich(I_rotated,sigma_level = 5)
        # Q_rotated = RemoveGlich(Q_rotated,sigma_level = 5)  
        
        if method == 'welch':
            freqnoise, Sxx = sps.welch(I_rotated,fs = fs,window = 'hann',nperseg = nperseg)
        
        elif method == "periodogram":
            freqnoise, Sxx = sps.periodogram(I_rotated,fs = fs,window = 'hann',nfft = nperseg)
        
        if combinespectrum == True:
            freqnoise,Sxx = CombineSpectrum(freqnoise,Sxx);
        
        
        del I_rotated
        
        if method == 'welch':
        
            freqnoise, Syy = sps.welch(Q_rotated,fs = fs,window = 'hann',nperseg = nperseg)
        
        elif method == 'periodogram':
            
            freqnoise, Syy = sps.periodogram(Q_rotated,fs = fs,window = 'hann',nfft = nperseg)
        
        
        
        if combinespectrum == True:
            freqnoise,Sxx = CombineSpectrum(freqnoise,Syy)
        
        del Q_rotated
        
        if onres == True:
            self.freqOnResNoise = freqnoise;
            self.SxxOnRes = Sxx;
            self.SyyOnRes = Syy;
            # self.ZNoiseOnRes = noisedata[0:1000]
            # self.QNoiseOnRes = Q_rotated[0:1000]
            self.TNoiseOnRes = np.arange(0,1000,1)*1/fs
        else:
            self.freqOffResNoise = freqnoise
            self.SxxOffRes = Sxx;
            self.SyyOffRes = Syy;
            # self.INoiseOffRes = I_roatated[0:1000]
            # self.ZNoiseOffRes = noisedata[0:1000]
            self.TNoiseOffRes = np.arange(0,1000,1)*1/fs
            
        # self.resultIBaseline = resultIBaseline
        # self.resultQBaseline = resultQBaseline
    
    def cal_pulse(self,pulsedata,pulsefreqmeas,pulsewidth = 50,fs = 5e6):
        
        # if self.hasnoisedata == True:
        indx = np.argmin(np.abs(self.freq-pulsefreqmeas))
        indx = indx
        
        #the delay of the transmission 
        baselinenoise = self.resultIBaseline[indx] + 1j*self.resultQBaseline[indx]
        
        cal_factor = (self.I[indx] + 1j*self.Q[indx])/(self.INorm[indx] + 1j*self.QNorm[indx])
        # baseline = self.resultIBaseline + 1j*self.resultQBaseline
        
        # f_interp_baseline = interp.interp1d(self.freq,baseline)
        
        # baselinenoise = f_interp_baseline(pulsefreqmeas)
        
        # f_interp_circle = interp.interp1d(self.freq,self.INorm+1j*self.QNorm)
        
        # r_base = f_interp_circle(pulsefreqmeas)
        
        rcal = np.abs(self.INorm[indx] + 1j*self.QNorm[indx] - (self.x0+1j*self.y0))
        # rcal = np.abs(r_base - (self.x0+1j*self.y0))
        
        #calibrate the noise data
        pulsedata = pulsedata/cal_factor
        
        pulse_amp = np.abs(pulsedata-(self.x0+1j*self.y0))/rcal
        
        pulse_phase = np.arctan2(np.imag(pulsedata)-self.y0,np.real(pulsedata)-self.x0)
        
        self.pulseWidths.append(pulsewidth)
        self.pulseAmps.append(pulse_amp)
        self.pulsePhases.append(pulse_phase)
        self.pulseTimes.append(np.arange(0,len(pulse_phase))*1/fs)
        self.pulseIs.append(np.real(pulsedata))
        self.pulseQs.append(np.imag(pulsedata))
        
    def cal_pulse_ER(self,Is,Qs,pulsefreqmeas,pulsewidth = 50,fs = 5e6,cleardata = False):
        
        indx = np.argmin(np.abs(self.freq-pulsefreqmeas))
        
        cal_factor = (self.I[indx] + 1j*self.Q[indx])/(self.INorm[indx] + 1j*self.QNorm[indx])
        
        rcal = np.abs(self.INorm[indx] + 1j*self.QNorm[indx] - (self.x0+1j*self.y0))
        # rcal = np.abs(r_base - (self.x0+1j*self.y0))
        
        #calibrate the noise data
        
        if cleardata == True:
            self.pulseERIs = []
            self.pulseERQs = []
            
        
        for indx, I in enumerate(Is):
            
            Q = Qs[indx]
            
            pulsedata = I + 1j*Q
            
            pulsedata = pulsedata/cal_factor
        
            pulse_amp = np.abs(pulsedata-(self.x0+1j*self.y0))/rcal
        
            pulse_phase = np.arctan2(np.imag(pulsedata)-self.y0,np.real(pulsedata)-self.x0)
        
            self.pulseERIs.append(np.real(pulsedata))
            self.pulseERQs.append(np.imag(pulsedata))
            
            
        
        
            
        return pulse_amp,pulse_phase
        
        
        
    def cal_pulse_outside(self,pulsedata,pulsefreqmeas,pulsewidth = 50,fs = 5e6):
        
        # if self.hasnoisedata == True:
        indx = np.argmin(np.abs(self.freq-pulsefreqmeas))
        indx = indx
        
        #the delay of the transmission 
        baselinenoise = self.resultIBaseline[indx] + 1j*self.resultQBaseline[indx]
        
        cal_factor = (self.I[indx] + 1j*self.Q[indx])/(self.INorm[indx] + 1j*self.QNorm[indx])
        # baseline = self.resultIBaseline + 1j*self.resultQBaseline
        
        # f_interp_baseline = interp.interp1d(self.freq,baseline)
        
        # baselinenoise = f_interp_baseline(pulsefreqmeas)
        
        # f_interp_circle = interp.interp1d(self.freq,self.INorm+1j*self.QNorm)
        
        # r_base = f_interp_circle(pulsefreqmeas)
        
        rcal = np.abs(self.INorm[indx] + 1j*self.QNorm[indx] - (self.x0+1j*self.y0))
        # rcal = np.abs(r_base - (self.x0+1j*self.y0))
        
        #calibrate the noise data
        pulsedata = pulsedata/cal_factor
        
        pulse_amp = np.abs(pulsedata-(self.x0+1j*self.y0))/rcal
        
        pulse_phase = np.arctan2(np.imag(pulsedata)-self.y0,np.real(pulsedata)-self.x0)
        
        return pulse_amp,pulse_phase

    def torch_params(self):
        """Reset params attribute to ``None``."""
        self.params = None
        self.hasParams = False

    def do_lmfit(self, fitFn, label='default', fit_type='IQ', **kwargs):
        r"""Run lmfit on an existing resonator object and update the results.

        Parameters
        ----------
        fitFn : function
            fitFn must have the signature fitFn(params, res, residual, **kwargs).
            If residual == True, fitFn must return a 1D list-like object of
            residuals with form [I residual, Q residual] where [A, B] means
            concatenate. Otherwise it must return the model data in the same form.

        label: string
            A label to use as a key when storing results from the fit to the
            lmfit_results dict.

        fit_type: string
            Indicates the type of fit to be run. For some types of fits, certain
            quantities will automatically be calculated and added to the resonator
            object. For instance, 'IQ' will cause the magnitude, phase, I, and Q
            as well as associated residuals to be calculated.

        kwargs : optional keywords
            Use this to override any of the lmfit parameter initial guesses or
            toggle whether the paramter varys. Example: ``qi=1e6`` is equivalent
            to calling ``Resonator.params['qi'].value = 1e6``. Example:
            ``qi_vary=False`` will set ``Resonator.params['qi'].vary = False``.
            Any parameter name can be used in this way.
        """
        assert self.hasParams == True, "Must load params before running a fit."

        #Update any of the default Parameter guesses
        if kwargs is not None:
            for key, val in kwargs.items():
                #Allow for turning on and off parameter variation
                if '_vary' in key:
                    key = key.split('_')[0]
                    if key in self.params.keys():
                        if (val is True) or (val is False):
                            self.params[key].vary = val
                elif key in self.params.keys():
                    self.params[key].value = val
                else:
                    raise ValueError("Unknown key: "+key)

        # #Make complex vectors of the form cData = [reData, imData]
        # cmplxData = np.concatenate((self.I, self.Q), axis=0)

        # if (self.sigmaI is not None) and (self.sigmaQ is not None):
        #     cmplxSigma = np.concatenate((self.sigmaI, self.sigmaQ), axis=0)
        # else:
        #     cmplxSigma = None

        # #Create a lmfit minimizer object
        # minObj = lf.Minimizer(fitFn, self.params, fcn_args=(self.freq, cmplxData, cmplxSigma))

        #Create a lmfit minimizer object
        minObj = lf.Minimizer(fitFn, self.params, fcn_args=(self, True),nan_policy = 'propagate')

        lmfit_result = minObj.minimize(method = 'leastsq')

        #Call the lmfit minimizer method and minimize the residual
        if self.lmfit_result is None:
            self.lmfit_result = {}

        self.lmfit_result[label] = {}
        self.lmfit_result[label]['fit_type'] = fit_type
        self.lmfit_result[label]['result'] = lmfit_result
        self.lmfit_result[label]['values'] = np.asarray([val.value for key, val in lmfit_result.params.items() if val.vary is True])
        self.lmfit_result[label]['labels'] = [key for key, val in lmfit_result.params.items() if val.vary is True]

        #NOTE: These are likely to be deprecated
        if label == 'default':
            self.lmfit_vals = self.lmfit_result[label]['values']
            self.lmfit_labels = self.lmfit_result[label]['labels']


        #Set the hasFit flag NOTE:(scheduled for deprecation)
        self.hasFit = True

        #NOTE: This whole block may be deprecated
        if (fit_type == 'IQ') and (label == 'default'):
            #Add the data back to the final minimized residual to get the final fit
            #Also calculate all relevant curves
            cmplxResult = fitFn(self.lmfit_result[label]['result'].params, self, residual=False)
            
            
            cmplxResidual = self.lmfit_result[label]['result'].residual

            #Split the complex data back up into real and imaginary parts
            residualI, residualQ = np.split(cmplxResidual, 2)
            
            resultI, resultQ = np.split(cmplxResult, 2)
            

            resultMag = np.abs(resultI + 1j*resultQ)
            resultPhase = np.arctan2(resultQ,resultI)
            
            #####the following is modified by Jie Hu for noise calculation 
            
            cmplxBaseline = fitFn(self.lmfit_result[label]['result'].params, 
                                  self, residual=False,only_baseline = True)
            resultIBaseline, resultQBaseline = np.split(cmplxBaseline, 2)
            
            resultZnorm = (resultI + 1j* resultQ)/(resultIBaseline + 1j*resultQBaseline)
            resultINorm, resultQNorm = [np.real(resultZnorm),np.imag(resultZnorm)]
            
            resultMagNorm = np.abs(resultZnorm)
            resultPhaseNorm = np.arctan2(resultQNorm,resultINorm)
            
            Znorm = (self.I + 1j*self.Q)/(resultIBaseline + 1j*resultQBaseline)
            INorm,QNorm = [np.real(Znorm),np.imag(Znorm)]
            
            MagNorm = np.abs(Znorm)
            PhaseNorm = np.arctan2(QNorm,INorm)
            
            x0,y0,r0 = calcirclecenter(resultINorm,resultQNorm)
            
            f0 = self.lmfit_result['default']['values'][1]
            
            indx = np.argmin(np.abs(self.freq-f0))
            
            INormRes = INorm[indx]
            QNormRes = QNorm[indx]
            ZNormRes = INormRes + 1j*QNormRes
            
            
            #####end of the modification  
            
            
            #Add some results back to the resonator object
            self.residualI = residualI
            self.residualQ = residualQ
            self.resultI = resultI
            self.resultQ = resultQ
            self.resultMag = resultMag
            self.resultPhase = resultPhase
            self.resultIBaseline = resultIBaseline
            self.resultQBaseline = resultQBaseline
            
            #####the following is modified by Jie Hu for noise calculation
            self.resultINorm = resultINorm
            self.resultQNorm = resultQNorm
            self.resultMagNorm = resultMagNorm
            self.resultPhaseNorm = resultPhaseNorm
        
            
            self.f0 = self.lmfit_result['default']['values'][1]
            self.Qc = self.lmfit_result['default']['values'][2]
            self.Qi = self.lmfit_result['default']['values'][3]
            
            self.INorm = INorm
            self.QNorm = QNorm
            self.MagNorm = MagNorm
            self.PhaseNorm = PhaseNorm
            self.x0 = x0;
            self.y0 = y0;
            self.r0 = r0;
            
            self.INormRes = INormRes
            self.QNormRes = QNormRes
            self.ZNormRes = ZNormRes
            
            
            #####end of the modification
            
        if (fit_type == 'IQ') and (label == 'sevenparams'):
            #Add the data back to the final minimized residual to get the final fit
            #Also calculate all relevant curves
            cmplxResult = fitFn(self.lmfit_result[label]['result'].params, self, residual=False)
            
            
            cmplxResidual = self.lmfit_result[label]['result'].residual

            #Split the complex data back up into real and imaginary parts
            residualI, residualQ = np.split(cmplxResidual, 2)
            
            resultI, resultQ = np.split(cmplxResult, 2)
            

            resultMag = np.abs(resultI + 1j*resultQ)
            resultPhase = np.arctan2(resultQ,resultI)
            # print(res.df)
            #####the following is modified by Jie Hu for noise calculation 
            
            cmplxBaseline = fitFn(self.lmfit_result[label]['result'].params, 
                                  self, residual=False,only_baseline = True)
            resultIBaseline, resultQBaseline = np.split(cmplxBaseline, 2)
            
            resultZnorm = (resultI + 1j* resultQ)/(resultIBaseline + 1j*resultQBaseline)
            resultINorm, resultQNorm = [np.real(resultZnorm),np.imag(resultZnorm)]
            
            resultMagNorm = np.abs(resultZnorm)
            resultPhaseNorm = np.arctan2(resultQNorm,resultINorm)
            
            Znorm = (self.I + 1j*self.Q)/(resultIBaseline + 1j*resultQBaseline)
            INorm,QNorm = [np.real(Znorm),np.imag(Znorm)]
            
            MagNorm = np.abs(Znorm)
            PhaseNorm = np.arctan2(QNorm,INorm)
            
            x0,y0,r0 = calcirclecenter(resultINorm,resultQNorm)
            
            f0 = self.lmfit_result[label]['values'][0]
            
            indx = np.argmin(np.abs(self.freq-f0))
            
            INormRes = INorm[indx]
            QNormRes = QNorm[indx]
            ZNormRes = INormRes + 1j*QNormRes
            
            
            #####end of the modification  
            
            
            #Add some results back to the resonator object
            self.residualI = residualI
            self.residualQ = residualQ
            self.resultI = resultI
            self.resultQ = resultQ
            self.resultMag = resultMag
            self.resultPhase = resultPhase
            self.resultIBaseline = resultIBaseline
            self.resultQBaseline = resultQBaseline
            
            #####the following is modified by Jie Hu for noise calculation
            self.resultINorm = resultINorm
            self.resultQNorm = resultQNorm
            self.resultMagNorm = resultMagNorm
            self.resultPhaseNorm = resultPhaseNorm
        
            
            self.f0 = self.lmfit_result[label]['values'][0]
            self.Qc = self.lmfit_result[label]['values'][2]
            self.Qr = self.lmfit_result[label]['values'][1]
            self.phi0 = self.lmfit_result[label]['values'][3]
            self.Qi = 1/(1/self.Qr - 1/self.Qc)
            
            complQc = self.Qc*np.exp(1j*(-self.phi0))
            Qc = 1/(1/complQc).real
            # absQc = Ql/(2.*r0)
            # complQc = absQc*np.exp(1j*((-1.)*phi0))
            # Qc = 1./(1./complQc).real    # here, taking the real part of (1/complQc) from diameter correction method
            # Qi_dia_corr = 1./(1./Ql-1./Qc)
            # Qi_no_corr = 1./(1./Ql-1./absQc)
            self.Qi_dia_corr = 1./(1./self.Qr-1./Qc)
            
            
            
            
            self.INorm = INorm
            self.QNorm = QNorm
            self.MagNorm = MagNorm
            self.PhaseNorm = PhaseNorm
            self.x0 = x0;
            self.y0 = y0;
            self.r0 = r0;
            
            self.INormRes = INormRes
            self.QNormRes = QNormRes
            self.ZNormRes = ZNormRes
            
            
            #####end of the modification

    def torch_lmfit(self, label='default'):
        r"""Reset all the lmfit attributes to ``None`` and set ``hasFit = False``.

        Parameters
        ----------
        label : string (optional)
            Choose which fit to kill off.

        Return
        ------
        deleted_fit : dict or None
            Return the fit that was deleted or None
        """

        deleted_fit = None

        if self.lmfit_result is not None:
            if label in self.lmfit_result.keys():
                deleted_fit = self.lmfit_result.pop(label)

                if label == 'default':
                    self.lmfit_vals = None
                    self.lmfit_labels = None
            
                if (deleted_fit['fit_type'] == 'IQ') and label == 'default':
                    
                    self.residualI = None
                    self.residualQ = None
                    self.resultI = None
                    self.resultQ = None
                    self.resultMag = None
                    self.resultPhase = None


                if len(self.lmfit_result.keys()) == 0:
                    self.lmfit_result = None
                    self.hasFit = False

        return deleted_fit


    def do_emcee(self, fitFn, label='default', **kwargs):
        r"""Run the Monte-Carlo Markov Chain routine to generate samples for
        each parameter given a model.

        Parameters
        ----------
        fitFn : function
            fitFn must have the signature fitFn(params, res, residual, **kwargs).
            If residual == True, fitFn must return a 1D list-like object of
            residuals with form [I residual, Q residual] where [A, B] means
            concatenate. Otherwise it must return the model data in the same form.

        label : string (optional)
            A label to assign to the fit results. This will be the dict key they
            are stored under in the emcee_results dict. Also, if label matches a
            label in lmfit_results, then that params object will be used to seed
            the emcee fit.

        kwargs : optional keyword arguments
            These are passed through to the ``lmfit.Minimizer.emcee`` method.
            See the ``lmfit`` documentation for more information.

        """
        #Should do the following (have not implemented any of this yet):
        #Pack MLE values into their own params object by adding back in non-varying Parameters
        #Should consider the ability to filter results for better parameter estimations
        #Probably should make a nice easy output to the corner Package
        #Smart way to add in error parameter as nuisance without breaking auto-guessing

        #minimizerObj.emcee already updates parameters object to result
        #This means can call res.emcee_result.params to get results

        #Create a lmfit minimizer object
        if self.hasFit:
            if self.lmfit_result is not None:
                if label in self.lmfit_result.keys():
                    emcee_params = self.lmfit_result[label]['result'].params
        else:
            assert self.hasParams == True, "Must load params before running emcee."
            emcee_params = self.params

        minObj = lf.Minimizer(fitFn, emcee_params, fcn_args=(self, True))

        #Run the emcee and add the result in
        emcee_result = minObj.emcee(**kwargs)

        if self.emcee_result is None:
            self.emcee_result = {}

        self.emcee_result[label] = {}
        self.emcee_result[label]['result'] = emcee_result
        
        #Get the emcee 50th percentile data and uncertainties at 16th and 84th percentiles
        emcee_vals = np.asarray([np.percentile(emcee_result.flatchain[key], 50) for key in emcee_result.flatchain.keys()])
        err_plus = np.asarray([np.percentile(emcee_result.flatchain[key], 84) for key in emcee_result.flatchain.keys()])
        err_minus = np.asarray([np.percentile(emcee_result.flatchain[key], 16) for key in emcee_result.flatchain.keys()])

        #Pack these values into the fit storage dict
        self.emcee_result[label]['values'] = emcee_vals

        #Make a list of tuples that are (+err, -err) for each paramter
        self.emcee_result[label]['emcee_sigmas'] = list(zip(err_plus-emcee_vals, emcee_vals-err_minus))

        #It is also useful to have easy access to the maximum-liklihood estimates
        self.emcee_result[label]['mle_vals'] = emcee_result.flatchain.iloc[np.argmax(emcee_result.lnprob)]

        #This is useful because only varying parameters have mle vals
        self.emcee_result[label]['mle_labels'] = self.emcee_result[label]['mle_vals'].keys()


        if label == 'default':
            self.emcee_vals = self.emcee_result[label]['values']

            #Make a list of tuples that are (+err, -err) for each paramter
            self.emcee_sigmas = self.emcee_result[label]['emcee_sigmas']

            #It is also useful to have easy access to the maximum-liklihood estimates
            self.mle_vals = self.emcee_result[label]['mle_vals']

            #This is useful because only varying parameters have mle vals
            self.mle_labels = self.emcee_result[label]['mle_labels']

            #This is also nice to have explicitly for passing to triangle-plotting routines
            self.chain = emcee_result.flatchain.copy()
        
        self.hasChain = True

    def burn_flatchain(self, num_samples=0, label='default'):
        r"""Burns off num_samples samples from each of the chains and then reflattens. Recalculates all
        statistical quantities associated with the emcee run and saves them under the original
        label, but with the suffix '_burn' appended to the various keys. Does not modify original chain."""
        
        flatchain_with_burn = pd.DataFrame()
        chains = self.emcee_result[label]['result'].chain
        
        for ix, chain in enumerate(chains.T):
            flatchain_with_burn[self.emcee_result[label]['mle_labels'][ix]] = chain[num_samples:].flat

        #Get the emcee 50th percentile data and uncertainties at 16th and 84th percentiles
        emcee_vals = np.asarray([np.percentile(flatchain_with_burn[key], 50) for key in flatchain_with_burn.keys()])
        err_plus = np.asarray([np.percentile(flatchain_with_burn[key], 84) for key in flatchain_with_burn.keys()])
        err_minus = np.asarray([np.percentile(flatchain_with_burn[key], 16) for key in flatchain_with_burn.keys()])

        #Pack these values into the fit storage dict with suffix _burn
        self.emcee_result[label]['values_burn'] = emcee_vals

        #Make a list of tuples that are (+err, -err) for each paramter
        self.emcee_result[label]['emcee_sigmas_burn'] = list(zip(err_plus-emcee_vals, emcee_vals-err_minus))

        #TODO: Implement this!
        #It is also useful to have easy access to the maximum-liklihood estimates
        #self.emcee_result[label]['mle_vals_burn'] = flatchain_with_burn.iloc[np.argmax(emcee_result.lnprob)]

        #Add the burned flatchain in its own key
        self.emcee_result[label]['flatchain_burn'] = flatchain_with_burn

    def torch_emcee(self, label='default'):
        r"""Set the emcee-related attributes to ``None`` and ``hasChain = False``.
        Parameters
        ----------
        label : string (optional)
            Which fit to torch
            
        Return
        ------
        deleted_fit : dict
            The fit that is deleted is returned, or None."""

        deleted_fit = None
        
        if self.emcee_result is not None:
            if label in self.emcee_result.keys():
                deleted_fit = self.emcee_result.pop(label)

                if label == 'default':
                    self.emcee_vals = None
                    self.emcee_sigmas = None
                    self.mle_vals = None
                    self.mle_labels = None
                    self.chain = None


                if len(self.emcee_result.keys()) == 0:
                    self.hasChain = False
                    self.emcee_result = None
        
        return deleted_fit

#This creates a resonator object from a data dictionary. Optionally performs a fit, and
#adds the fit data back in to the resonator object

def CombineSpectrum(freqnoise,Sxx,freqsegs = [50,1e3,1e4,1e5,1e6,1e7] ,respoints = [1,10,1e2,1e3,1e4,1e5]):
    
    #this function requires that the spectrum resolution to be smaller than 1Hz
    
    df = freqnoise[1]-freqnoise[0]
    
    for i, freqseg in enumerate(freqsegs):
        
        sumpoints = round(respoints[i]/df);
        
        segpoints = freqsegs
        
    



def RemoveGlich(data,sigma_level = 5):
    
    # assume that the data is normal distributed;
    
    # data_amp = np.abs(data)
    
    mean = np.mean(data)           
    
    std = np.std(data)
    
    value_smaller = np.abs(data - mean) < sigma_level*std
    
    print(np.sum(1-value_smaller))
    
    value_smaller = value_smaller*data + (1-value_smaller) * mean
    
    return value_smaller

def calcirclecenter(resultINorm,resultQNorm):
    
    lendata = len(resultINorm)
    
    x1 = resultINorm[0]
    y1 = resultQNorm[0]
    
    x2 = resultINorm[int(lendata/3)]
    y2 = resultQNorm[int(lendata/3)]
    
    x3 = resultINorm[int(lendata/3)*2]
    y3 = resultQNorm[int(lendata/3)*2]
    
    # x0 = (x1 + x2 + x3)/3
    # y0 = (y1 + y2 + y3)/3
    
    a = x1-x2;
    b = y1-y2;
    c = x1-x3;
    d = y1-y3;
    
    e = (x1*x1 - x2*x2 + y1*y1 - y2*y2)/2
    f = (x1*x1 - x3*x3 + y1*y1 - y3*y3)/2;
    
    det = b*c - a*d;
    
    if det <1e-5:
        radius = -1;
        x0 = 0;
        y0 = 0;
        
    else:
        x0 = -(d*e-b*f)/det;
        y0 = -(a*f-c*e)/det;
        
        radius = np.sqrt((x1-x0)**2 + (y1-y0)**2)
    return x0,y0,radius;
    
    
    
    
    # return x0,y0,r0

def makeResFromData(dataDict, paramsFn = None, fitFn = None, fitFn_kwargs=None, paramsFn_kwargs=None):
    """Create a Resonator object from a data dictionary.

    Parameters
    ----------
    dataDict : dict
        Must have the following keys: 'I', 'Q', 'temp', 'pwr', 'freq', 'name'.
        Optional keys are: 'sigmaI', 'sigmaQ'

    paramsFn : function (optional)
        A function that initializes and returns an lmfit parameters object for
        passing to fitFn.

    fitFn : function (optional)
        If a fit function is passed, an lmfit minimization will be done
        automatically.

    fitFn_kwargs : dict (optional)
        A dict of keyword arguments passed to fitFn.

    paramsFn_kwargs: dict (optional)
        A dict of keyword arguments passed to paramsFn.

    Returns
    -------
    res : ``Resonator`` object or ``None``
        A Resonator object or ``None`` if there is an error loading the data.

    """
    if fitFn is not None:
        assert paramsFn is not None, "Cannot pass a fitFn without also passing a paramsFn"

    #Check dataDict for validity
    expectedKeys = ['name', 'temp', 'pwr', 'freq', 'I', 'Q']
    assert all(key in dataDict.keys() for key in expectedKeys), "Your dataDict is missing one or more keys"

    resName = dataDict['name']
    temp = dataDict['temp']
    pwr = dataDict['pwr']
    freqData = dataDict['freq']
    IData = dataDict['I']
    QData = dataDict['Q']

    #Process the optional keys
    if 'sigmaI' in dataDict.keys():
        sigmaI = dataDict['sigmaI']
    else:
        sigmaI = None

    if 'sigmaQ' in dataDict.keys():
        sigmaQ = dataDict['sigmaQ']
    else:
        sigmaQ = None

    #create Resonator object
    res = Resonator(resName, temp, pwr, freqData, IData, QData, sigmaI, sigmaQ)

    #Process the fit parameters
    if paramsFn is not None:
        if paramsFn_kwargs is not None:
            res.load_params(paramsFn, **paramsFn_kwargs)
        else:
            res.load_params(paramsFn)

    #Run a fit on the resonator if a fit function is specified
    if fitFn is not None:
        if fitFn_kwargs is not None:
            res.do_lmfit(fitFn, **fitFn_kwargs)
        else:
            res.do_lmfit(fitFn)

    #Return resonator object
    return res

def makeResList(fileFunc, dataPath, resName, **fileFunc_kwargs):
    """Create a list of resonator objects from a directory of dataDict

    Parameters
    ----------
    fileFunc : function
        A function that converts a single data file into a dictionary. The
        resulting dictionary must have the following keys: 'I', 'Q', 'temp',
        'pwr', 'freq', 'name', and may have the following ptional keys:
        'sigmaI', 'sigmaQ'

    dataPath : string
        Path to the directory containing the data files that will be processed
        by fileFunc.

    resName : string
        The name of your resonator. This can be anything, but it is useful to
        use the same name for every data file that comes from the same physical
        resonator.

    fileFunc_kwargs : dict
        Keyword arguments to pass through to the fileFunc

    """
    #Find the files that match the resonator you care about
    fileList = glob.glob(dataPath + resName + '_*')

    #loop through files and process all the data
    fileDataDicts = []

    for f in fileList:
        fileDataDicts.append(fileFunc(f, **fileFunc_kwargs))

    #Create resonator objects from the data 
    #makeResFromData returns a tuple of (res, temp, pwr),
    #but only care about the first one
    resList = [makeResFromData(fileDataDict) for fileDataDict in fileDataDicts]

    return resList

#Index a list of resonator objects easily
def indexResList(resList, temp=None, pwr=None, **kwargs):
    """Index resList by temp and pwr.

    Parameters
    ----------
    resList : list-like
        resList is a list of ``scraps.Resonator`` objects
    temp : numeric
        The temperature of a single Resonator object.
    pwr : int
        The power of a single Resonator object

    itemp : boolean (optional)
        Switch to determine whether lookup uses temp or itemp (rounded value of
        temp). Default is ``False``.

    Returns
    -------
    index : int or list
        Index is the index of the Resonator in resList or a list of indices of
        all matches if only pwr or only temp is specified.

    Notes
    -----
    indexResList does not check for duplicates and will return the first match.

    """
    itemp = kwargs.pop('itemp', False)
    assert itemp in [True, False], "'itemp' must be boolean."

    assert (pwr is not None) or (temp is not None), "Must specify at least either a temp or a pwr."


    if (pwr is not None) and (temp is not None):
        for index, res in enumerate(resList):
            if itemp is True:
                if res.itemp == temp and res.pwr == pwr:
                    return index
            else:
                if np.isclose(res.temp, temp) and res.pwr == pwr:
                    return index
    elif (pwr is None):
        index = []
        for ix, res in enumerate(resList):
            if itemp is True:
                if res.itemp == temp:
                    index.append(ix)
                else:
                    if np.isclose(res.temp, temp):
                        index.append(ix)
    elif (temp is None):
        index = []
        for ix, res in enumerate(resList):
            if res.pwr == pwr:
                index.append(ix)
        return index

    return None

def print_resList(resList):
    """Print all the temperatures and powers in a table-like form"""
    #Get all possible powers
    pwrs = np.unique([res.pwr for res in resList])

    #This will hold a list of temps at each power
    tlists = []
    max_len = 0

    #Populate the lists of temps for each power
    for p in pwrs:
        tlist = [res.temp for res in resList if res.pwr == p]
        tlist.sort()
        tlists.append(tlist)
        if len(tlist) > max_len:
            max_len = len(tlist)

    for ix, tlist in enumerate(tlists):
        pad = max_len - len(tlist)
        tlist = tlist + pad*['NaN']
        tlists[ix] = tlist

    block = zip(*tlists)

    print(repr(list(pwrs)).replace(',', ',\t'))
    for b in block:
        print(repr(b).replace(',', ',\t'))



def block_check_resList(resList, sdev=0.005, prune=False, verbose=True):
    """Helper tool for preparing a resList with missing data for resSweep"""
    #Get all possible powers
    pwrs = np.unique([res.pwr for res in resList])

    #This will hold a list of temps at each power
    tlists = []

    #Populate the lists of temps for each power
    for p in pwrs:
        tlist = [res.temp for res in resList if res.pwr == p]
        tlist.sort()
        tlists.append(tlist)

    #Calculate the lengths and find the shortest one
    lens = [len(tl) for tl in tlists]
    shortest = min(lens)

    if all(el == shortest for el in lens) and verbose:
        print('All lists have same length.')
    else:
        print('Lengths for each set of powers: ',list(zip(pwrs,lens)))

    #Zip the lists into tuples and take the standard deviation
    #of each tuple. All the elements in each tuple should be
    #nominally the same, so the stdev should be small unless
    #one of the elements doesn't match. Return the first
    #instance of the stdev being too high
    block = list(zip(*tlists))
    bad_ix = np.argmax([np.std(x) > sdev for x in block])

    #If the first row is returned, everything could be ok. Check first row.
    if bad_ix == 0:
        if np.std(block[0]) < sdev:
            bad_ix = -1

    if verbose:
        print("Bad index: ", bad_ix)

    if bad_ix >= 0:

        if verbose:
            for i in np.arange(-2,3):
                if (bad_ix+i < len(block)) and (bad_ix+i >= 0):
                    print(repr(block[bad_ix+i]).replace(',', ',\t'))
                    block_ixs = []
                    for block_ix, block_temp in enumerate(block[bad_ix+i]):
                        block_ixs.append(indexResList(resList, block_temp, pwrs[block_ix]))
                    print(repr(block_ixs).replace(',', ',\t'))


            #The longer list is where the extra file is most likely
            #so return the temp, power, and resList index of the
            #suspect.
            for i, x in enumerate(block[bad_ix]):
                if np.abs(x-np.mean(block[bad_ix])) > np.std(block[bad_ix]):

                    tl = tlists[i]
                    t = tl[bad_ix]
                    p = pwrs[i]
                    res_ix = indexResList(resList, t, p)
                    if verbose:
                        print('T=',t, 'P=',p, 'Res index=',res_ix)
                    if prune:
                        resList.pop(res_ix)





import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import scipy.interpolate as interp;
import lmfit as lm



def fit_sin(t,A,phi,I0,f):

    return A*np.sin(2*np.pi*f*t + phi) + I0

def fit_cos(t,A,phi,I0,f):
    
    return A*np.cos(2*np.pi*f*t + phi) + I0

def IQ_binary_old(filename):
    
    data = np.fromfile(filename,dtype=float,count = -1,sep = '')
    data_len = int((len(data)-1)/2)
    
    time = data[0]*np.arange(0,data_len)
    I = data[1:data_len+1]
    Q = data[data_len+1:]
    return time,I, Q

def IQ_binary_new(filename):
    
    data = np.fromfile(filename,dtype=float,count = -1,sep = '')
    data_len = int(data[1])
    
    time = data[0]*np.arange(0,data_len)
    I = data[2:data_len+2]
    Q = data[data_len+2:]
    return time,I, Q

def GetAmpPhaseError(t,I,Q,IF_freq,isplotting = False):
    
    #take the I as an reference for calibration
  
    poptQ,pcovQ = curve_fit(lambda t,A,phi,Q0: fit_sin(t,A,phi,Q0,IF_freq),t,Q)

    poptI,pcovI = curve_fit(lambda t,A,phi,I0: fit_cos(t,A,phi,I0,IF_freq),t,I)

    AQ = poptQ[0]
    phiQ = poptQ[1]
    Q0 = poptQ[2]
    
    # print("AQ:",AQ)
    # print("APhi:",phiQ)

    AI = poptI[0]
    phiI = poptI[1]
    I0 = poptI[2]
    
    delta_amp = AI/AQ
    delta_phi = phiQ-phiI
    
    if isplotting == True:
        I_fit = fit_cos(t,AI,phiI,I0,IF_freq)
        Q_fit = fit_sin(t,AQ,phiQ,Q0,IF_freq)
        
        f,(ax1,ax2) = plt.subplots(2,1,sharex = True)
        ax1.plot(t*1e6,I,label = 'I Meas')
        ax1.plot(t*1e6,I_fit,label = 'I Fit')
        
        ax1.legend()
        
        plt.title("$\Delta A$:{0:.3f} $\Delta \phi$:{1:.3f}".format(delta_amp,delta_phi))
        
        # ax1.text()
        
        ax1.set_ylabel("I(V)",fontsize = 20)
        
        ax2.plot(t*1e6,Q,label = 'Q Meas')
        ax2.plot(t*1e6,Q_fit,label = 'Q Fit')
        
#        ax1.ylabel("Q(V)")
        ax2.set_xlabel("Time($\mu s$)",fontsize = 20)
        ax2.set_ylabel("Q(V)",fontsize = 20)
        ax2.legend()
        plt.tight_layout()
        
        
    
    return delta_amp,delta_phi,I0,Q0
    
    



def IQCal(I,Q,delta_A,delta_phi,I0 = 0,Q0 = 0):
    
    #calibrate the DC offset
    #I: the raw I data
    #Q: the raw Q data
    #delta_A: the radio of the amp of Q and I
    #delta_phi: the phase imbalance between I and Q
    #I0,Q0: the DC offset of I and Q
    
    I1 = I-I0;
    Q1 = Q-Q0;
    Q1 = Q1*delta_A;
   
    
#    phi = np.arctan2((Q1 - I1*np.cos(delta_phi))/I1/np.cos(delta_phi))
    phi = np.arctan2(Q1 - I1*np.sin(delta_phi),I1*np.cos(delta_phi) )
#    plt.plot(phi)
    
    
    amp = np.sqrt((I1**2 + Q1**2)/(np.cos(phi)**2+np.sin(phi+delta_phi)**2));
    
    I_cal = amp*np.cos(phi)
    Q_cal = amp*np.sin(phi)
    
    return I_cal,Q_cal
    
def IQ_CalS21FromFile(IQ_file,IQ_caldata=None):
    

    indx,freq,I,Q,I0,Q0 = np.loadtxt(IQ_file,unpack = True)
    # indx,freq,I_thru,Q_thru,I0_thru,Q0_thru = np.loadtxt(IQ_thrufile,unpack = True)
    
    
    freq_cal = IQ_caldata[:,0]
    amp_imbalance = IQ_caldata[:,1]
    phase_imbalance = IQ_caldata[:,2]

    f_amp = interp.interp1d(freq_cal,amp_imbalance)
    f_phase = interp.interp1d(freq_cal,phase_imbalance)

    delta_A = f_amp(freq)
    delta_phase = f_phase(freq)    

    I_cal,Q_cal =IQCal(I,Q,delta_A,delta_phase, I0 = I0,Q0 = Q0)
    
    
    return freq,I_cal,Q_cal,I0,Q0

def IQ_CalS21(freq,I,Q,I0,Q0,IQ_caldata=None):
    
    # indx,freq,I,Q,I0,Q0 = np.loadtxt(IQ_file,unpack = True)
    # indx,freq,I_thru,Q_thru,I0_thru,Q0_thru = np.loadtxt(IQ_thrufile,unpack = True)
    
    
    
    
    if IQ_caldata is not None:
        freq_cal = IQ_caldata[:,0]
        amp_imbalance = IQ_caldata[:,1]
        phase_imbalance = IQ_caldata[:,2]
    
        if np.max(freq) > np.max(freq_cal):
            print("IQ cal out of range no calibration is done")
            return I,Q,I0,Q0 
        

        f_amp = interp.interp1d(freq_cal,amp_imbalance)
        f_phase = interp.interp1d(freq_cal,phase_imbalance)

        delta_A = f_amp(freq)
        delta_phase = f_phase(freq)
        
    else:
        
        delta_A = np.ones_like(freq)
        delta_phase = np.zeros_like(freq)
        
        
    I_cal,Q_cal =IQCal(I,Q,delta_A,delta_phase, I0 = I0,Q0 = Q0)
    
    
    return I_cal,Q_cal,I0,Q0

def IQ_CalS21_VNA(IQ_file):
    
    data = np.loadtxt(IQ_file)
    
    freq = data[:,1]
    amp = data[:,2]
    phase = data[:,3]
    
    S21 = 10**(amp/20)*np.exp(1j*phase)
    
    I = np.real(S21)
    Q = np.imag(S21)
    
    return freq,I,Q
    

def IQ_CalNoise(I,Q,freq_noise,freq,I0,Q0,IQ_caldata):
    
    indx,delta_A,delta_phase,I0_noise,Q0_noise = IQ_noiseCaldata(freq_noise,freq,I0,Q0,IQ_caldata)
    
    ICal,QCal = IQCal(I,Q,delta_A,delta_phase,I0_noise,Q0_noise)
    
    return ICal,QCal
    

def IQ_noiseCaldata(freq_noise,freq,I0,Q0,IQ_caldata):
    
    
    freq_cal = IQ_caldata[:,0]
    amp_imbalance = IQ_caldata[:,1]
    phase_imbalance = IQ_caldata[:,2]

    f_amp = interp.interp1d(freq_cal,amp_imbalance)
    f_phase = interp.interp1d(freq_cal,phase_imbalance)

    delta_A = f_amp(freq_noise)
    delta_phase = f_phase(freq_noise)  
    
    indx = np.argmin(np.abs(freq-freq_noise))
    
    I0_noise = I0[indx]
    Q0_noise = Q0[indx]
    
    
    # I_cal,Q_cal =IQCal(I_noise,Q_noise,delta_A,delta_phase, I0 = I0,Q0 = Q0)

    return indx,delta_A,delta_phase,I0_noise,Q0_noise













import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import re
from scipy import signal
import scipy as sp
import Functions.IQCalibration as IQCal
import os
import pickle


#the following three function comes from Mazin's Lab*********************
def covariance_from_psd(psd, size=None, window=None, dt=1.):
    autocovariance = np.real(np.fft.irfft(psd / 2., window) / dt)  # divide by 2 for single sided PSD
    if size is not None:
        autocovariance = autocovariance[:size]
    covariance = sp.linalg.toeplitz(autocovariance)
    return covariance


def filter_cutoff(filter_, cutoff):
    """
    This function addresses a problem encountered when generating filters from
    oversampled data. The high frequency content of the filters can be
    artificially large due to poor estimation of the noise and template.

    In this case it is useful to remove frequencies above the cutoff from the
    filter. Only use this function when addressing the above issue and if the
    majority of the signal frequencies are < cutoff.

    It is best to avoid this procedure if possible since removing the high
    frequency content will artificially make the filter periodic, throws away
    some information in the signal, and may negatively influence some of the
    intended filter properties.
    """
    freq = np.fft.rfftfreq(filter_.shape[0], d=1)
    filter_fft = np.fft.rfft(filter_, axis=0)
    filter_fft[freq > cutoff, ...] = 0
    filter_ = np.fft.irfft(filter_fft, filter_.shape[0], axis=0)
    return filter_

def wiener(*args, **kwargs):
    """
    Create a filter that minimizes the chi squared statistic when aligned
    with a photon pulse.

    Args:
        template:  numpy.ndarray
            The template with which to construct the filter.
        psd:  numpy.ndarray
            The power spectral density of the noise.
        nwindow: integer
            The window size used to compute the PSD.
        nfilter: integer (optional)
            The number of taps to use in the filter. The default is to use
            the template size.
        cutoff: float (optional)
            Set the filter response to zero above this frequency (in units of
            1 / dt). If False, no cutoff is applied. The default is False.
        fft: boolean (optional)
            If True, the filter will be computed in the Fourier domain, which
            could be faster for very long filters but will also introduce
            assumptions about periodicity of the signal. In this case, the
            psd must be the same size as the filter Fourier transform
            (nfilter // 2 + 1 points). The default is False, and the filter is
            computed in the time domain.
        normalize: boolean (optional)
            If False, the template will not be normalized. The default is True
            and the template is normalized to a unit response.
    Returns:
        filter_: numpy.ndarray
            The computed wiener filter.
    """
    # collect inputs
    template, psd, nwindow = args[0], args[1], args[2]
    nfilter = kwargs.get("nfilter", len(template))
    cutoff = kwargs.get("cutoff", False)
    fft = kwargs.get("fft", False)
    normalize = kwargs.get("normalize", True)

    # need at least this long of a PSD
    if nwindow < nfilter:
        raise ValueError("The psd must be at least as long as the length of the FFT of the filter")

    # pad the template if it's shorter than the filter (linear ramp to avoid discontinuities)
    if nfilter > len(template):
        template = np.pad(template, (0, nfilter - len(template)), mode='linear_ramp')
    # truncate the template if it's longer than the filter
    elif nfilter < len(template):
        template = template[:nfilter]

    if fft:  # compute the filter in the frequency domain (introduces periodicity assumption, requires nwindow=nfilter)
        if nwindow != nfilter:
            raise ValueError("The psd must be exactly the length of the FFT of the filter to use the 'fft' method.")
        template_fft = np.fft.rfft(template)
        filter_ = np.fft.irfft(np.conj(template_fft) / psd, nwindow)  # must be same size else ValueError
        filter_ = np.roll(filter_, -1)  # roll to put the zero time index on the far right

    else:  # compute filter in the time domain (nfilter should be << nwindow for this method to be better than fft)
        covariance = covariance_from_psd(psd, size=nfilter, window=nwindow)
        filter_ = np.linalg.solve(covariance, template)[::-1]

    # remove high frequency filter content
    if cutoff:
        filter_ = filter_cutoff(filter_, cutoff)

    # normalize
    if normalize:
        filter_ /= -np.matmul(template, filter_[::-1])
    else:
        filter_ *= -1   # "-" to give negative pulse heights after filtering
    return filter_

#the above three function comes from Mazin's Lab*********************


# def GenWienerFilter(template)
def CalNoiseRaw(res,IQCaldata,isonres = True):
    
    noiseIs = []
    noiseQs = []
    
    freq = res.freq;
    I0 = res.I0;
    Q0 = res.Q0
    
    if isonres == True:
        noisearray = res.noiseUnscaled
    else:
        noisearray = res.noiseOffresUnscaled
    
    for noisedata in noisearray:
    # pulsefreq = res.pulseUnscaled
        noisefreq = noisedata['freq']
        voffset = noisedata['voffsets']
        vgains = noisedata['vgains']
        
        # print("pulsefreq:",pulsefreq)
        
        fsample = 1/noisedata['dt']
        I = noisedata['traces'][0]*vgains[0] - voffset[0]
        Q = noisedata['traces'][1]*vgains[1] - voffset[1]
        
        I,Q = IQCal.IQ_CalNoise(I,Q,noisefreq,freq/1e9,I0,Q0,IQCaldata)
        
        noiseIs.append(I)
        noiseQs.append(Q)
        
    noiseIs = np.concatenate(noiseIs)
    noiseQs = np.concatenate(noiseQs)
    
        
    return noiseIs,noiseQs,noisefreq,fsample

def CalNoiseHDf(res,hdfile,IQCaldata,isonres = True):
    
    #calibrate the noise data from a hdf file
    
    noiseIs = []
    noiseQs = []
    
    freq = res.freq;
    I0 = res.I0;
    Q0 = res.Q0
    
    if isonres == True:
        # noisearray = res.noiseUnscaled
        noisedata = hdfile['noise']
    else:
        # noisearray = res.noiseOffresUnscaled
        noisedata = hdfile['noiseoffres']
        
    segnum = noisedata.attrs['groupnum']
    
    noisefreq = noisedata.attrs['measfreq']/1e9
    
    voffsets = noisedata['voffsets'][...]
    vgains = noisedata['vgains'][...]
    
    for i in range(segnum):
    # pulsefreq = res.pulseUnscaled
        
        # print("pulsefreq:",pulsefreq)
        IQdata = noisedata['IQ%d'%(i)][...]
        
        # fsample = 1/noisedata['dt']
        I = IQdata[0]*vgains[0] - voffsets[0]
        Q = IQdata[1]*vgains[1] - voffsets[1]
        
        I,Q = IQCal.IQ_CalNoise(I,Q,noisefreq,freq/1e9,I0,Q0,IQCaldata)
        
        noiseIs.append(I)
        noiseQs.append(Q)
        
    noiseIs = np.concatenate(noiseIs)
    noiseQs = np.concatenate(noiseQs)
    
        
    return noiseIs,noiseQs

def CalPulseRaw(pulsedata,IQCaldata,freq,I0,Q0,ispulse = True):
    
    pulsefreq = pulsedata['freq']
    voffset = pulsedata['voffsets']
    vgains = pulsedata['vgains']
    
    # print("pulsefreq:",pulsefreq)
    
    fsample = 1/pulsedata['dt']
    I = pulsedata['traces'][0]*vgains[0] - voffset[0]
    Q = pulsedata['traces'][1]*vgains[1] - voffset[1]
    
    I,Q = IQCal.IQ_CalNoise(I,Q,pulsefreq,freq/1e9,I0,Q0,IQCaldata)
    
    return I,Q,pulsefreq,fsample

def CombineBins(freq,spectrum,bins = [50,1e3,1e4,1e5,1e6],resolutions = [1,10,100,1000,10000]):
    
    indx_bins = []
    
    indx_bins.append(0)
    
    for i,freqseg in enumerate(bins):
        
        indx_bin = np.argmin(np.abs(freq-freqseg))
        
        if indx_bin == len(freq): #check if the bin is the end of the freq
            
            if indx_bins[i-1] == indx_bin:
                
                break
            
        indx_bins.append(indx_bin)
        
        
    if indx_bins[-1] < len(freq)-1: #check if the bin is the end of the freq
        
        bins.append(freq[-1])
         
        indx_bins.append(len(freq))
        
        resolutions.append(resolutions[-1])
        
    if len(bins) > len(indx_bins):
        
        bins = bins[0:len(indx_bins)]
        resolutions = resolutions[0:len(indx_bins)]
    
    
    df = freq[1]-freq[0]
    
    NumBinned = 0
    
    NumBins = []
    
    for i, freqseg in enumerate(bins):
        
        resolution = resolutions[i]
        
        if df > resolutions[i]:
            
            resolution = df
        
        if i == 0:
            NumBinned = NumBinned + np.floor(freqseg/resolution)
            NumBins.append(np.floor(freqseg/resolution))
        else:
            NumBinned = NumBinned + np.floor((freqseg - bins[i-1])/resolution)
            NumBins.append(np.floor((freqseg - bins[i-1])/resolution))
        
        # NumBins.append(NumBinned)
            
    freq_binned = np.zeros(int(NumBinned))
    
    spectrum_binned = np.zeros(int(NumBinned))
        
    
    count = 0
    
    for i, freqseg in enumerate(bins):
        
        resolution = resolutions[i]
        
        indx = indx_bins[i]
            
        if df > resolutions[i]:
            
            resolution = df
            
        # Numbin = int(np.floor(freqseg / resolution))
        
        Numbin = int(NumBins[i]) - 1
            
        Bincombine = int(np.floor(resolution/df))
            
        for jN in range(Numbin):
                
             freq_binned[count] = freq[indx + jN*Bincombine] 
             spectrum_binned[count] = np.sum(spectrum[indx+jN*Bincombine:indx+(jN+1)*Bincombine])/Bincombine
             
             count = count + 1
             
             
             
        
    return freq_binned[0:count],spectrum_binned[0:count]    

def ExtractDatafromString(data_string):
    
    data_string = re.findall('-?\ *[0-9]+\.?[0-9]*(?:[Ee]\ *[+-]?\ *[0-9]+)?',data_string)
    
#    data_string = re.findall('-?\d\.?\d*[Ee][+\-]?\d+',data_string);
    
    data = [float(i) for i in data_string]
    
    return data

def IQ_binary(filename):
    
    data = np.fromfile(filename,dtype=float,count = -1,sep = '')
    data_len = int((len(data)-1)/2)
    
    
    # print(data[0])
    # print(data[1])
    
    time = data[0]*np.arange(0,data_len)
    
    if len(data)>data_len*2+1:
        I = data[2:data_len+2]
        Q = data[data_len+2:2*data_len+2]
    else:
        I = data[1:data_len+1]
        Q = data[data_len+1:2*data_len+1]

    # pulse_phase = pulse_phase[mask]

        
        
    # plt.figure()
    
    # plot()
    
    return time,I, Q

def Truncate_Pulse(I,Q,dt = 1,t_start = -np.inf,t_stop = np.inf):
    
    data_len = len(I)
    time = dt*np.arange(0,data_len)

    # mask = (time)
    mask = (time >= t_start) * (time <= t_stop)
        
    I = I[mask]
    Q = Q[mask]
    
    return I,Q

def IQBinaryIntCal(FileIntCal):
    
    file = open(FileIntCal,'r')
    Lines = file.readlines()
    
    sampleT = ExtractDatafromString(Lines[0])
    samplePoint = ExtractDatafromString(Lines[1])
    vgains = ExtractDatafromString(Lines[2])
    voffsets =  ExtractDatafromString(Lines[3])
    
    return sampleT[0],int(samplePoint[0]),vgains,voffsets
        
    

def IQBinaryInt(FileInt,FileIntCal):
    
    data = np.fromfile(FileInt,dtype = np.short,count = -1)
    
    data_len = int(len(data)/2)
    
    sampleT, samplePoint,vgains,voffsets = IQBinaryIntCal(FileIntCal)
    
    time = sampleT*np.arange(0,data_len)
    I = data[0:data_len]*vgains[0] - voffsets[0]
    Q = data[data_len:2*data_len]*vgains[1] - voffsets[1]
    
    return time,I,Q
    
def lowpassfiltering(I,sample_freq,Q = None,cutoff = None,filter_order = 10):
    
    if cutoff == None:
        cutoff = sample_freq*0.4
    # print("sample_freq:",sample_freq)
#    cutoff = cutoff/sample_freq
    sos = signal.butter(filter_order, cutoff , 'lowpass', fs = sample_freq,output='sos' )
    
    I = signal.sosfilt(sos,I)
    
    if Q is None:
        
        return I
    
    else:
        
        Q = signal.sosfilt(sos,Q)

        return I,Q
    
    return I

def AverageArray(time,array,chuncksize = 10):
    
    newarraysize = int(np.floor(len(array)/chuncksize));
    newarray = np.zeros(newarraysize)
    newtimearray = np.zeros(newarraysize)
    for i in range(newarraysize):
        newarray[i] = np.mean(array[i*chuncksize:(i+1)*chuncksize])
        newtimearray[i] = time[i*chuncksize]
        
    return newtimearray,newarray
             
def GenReslist(datafolder,select_by = "Temp",InAtten = 30,Temp = 50,IDString = ["_cal.obj"]):

    reslist = []
    resfreq = []
    temp = []
    resfilenames = []
    
    
    if select_by == "Temp":
        
        AttenString = "Inatten{0:.0f}".format(InAtten)
        
        IDString.append(AttenString)
        
        folders = os.listdir(datafolder)
        
        tempfolders = []
        
        #Get the temp folders
        for folder in folders:
            
            if "mK" in folder and os.path.isdir(datafolder + "//" + folder): 
                
                tempfolders.append(datafolder + "//" + folder)
                
        # print(tempfolders)
                
        #Get the res filenames        
        for folder in tempfolders:
            
            files = os.listdir(folder)
            
            
            
            for file in files:
                
                correct_file = True
                
                for string in IDString:
                
                    correct_file = correct_file*(string in file)
                
                if correct_file:
                    resfilenames.append(folder + "//" + file)
        
                
        if resfilenames != []:
            
            # print(resfilenames)
            for file in resfilenames:
            
                f = open(file,'rb')
                res = pickle.load(f)
                f.close()
                
                reslist.append(res)
                
                temp.append(res.temp)
                
                resfreq.append(res.lmfit_vals[1])
                
            indxs = np.argsort(np.array(temp))    
            
            temp = [temp[k] for k in indxs]
            resfreq = [resfreq[k] for k in indxs]
            reslist = [reslist[k] for k in indxs]
                
    return temp,resfreq,reslist                  
                

def GenReslistobj(datafolder,select_by = "temp",InAtten = 30,Temp = 50,
                  AttenStringFormat = "Inatten{0:.0f}",TempStringFormat = "Res{0:.0f}mK",IDString = [".obj"]):

    reslist = []
    resfreq = []
    temp = []
    resfilenames = []
    
    
    if select_by == "temp":

        AttenString = AttenStringFormat.format(InAtten)
        
        IDString.append(AttenString)
        
        folders = os.listdir(datafolder)
        
        tempfolders = []
        
        #Get the temp folders
        for folder in folders:
            
            if "mK" in folder and os.path.isdir(datafolder + "//" + folder): 
                
                tempfolders.append(datafolder + "//" + folder)
        #Get the res filenames        
        for folder in tempfolders:
            
            files = os.listdir(folder)
            
            
            
            for file in files:
                
                correct_file = True
                
                for string in IDString:
                
                    correct_file = correct_file*(string in file)
                
                if correct_file:
                    resfilenames.append(folder + "//" + file)
        
                
        if resfilenames != []:
            for file in resfilenames:
            
                f = open(file,'rb')
                res = pickle.load(f)
                f.close()
                
                reslist.append(res)
                
                temp.append(res.temp)
                
                resfreq.append(res.lmfit_vals[1])
                
            indxs = np.argsort(np.array(temp))    
            
            temp = [temp[k] for k in indxs]
            resfreq = [resfreq[k] for k in indxs]
            reslist = [reslist[k] for k in indxs]
    
    if select_by == "pwr":
        
        TempString = TempStringFormat.format(Temp)
        
        IDString.append(TempString)
        
        folders = os.listdir(datafolder)
        
        tempfolders = []
        
        #Get the temp folders
        for folder in folders:
            
            if "mK" in folder and os.path.isdir(datafolder + "//" + folder): 
                
                tempfolders.append(datafolder + "//" + folder)
        #Get the res filenames        
        for folder in tempfolders:
            
            files = os.listdir(folder)

            for file in files:
                
                correct_file = True
                
                for string in IDString:
                
                    correct_file = correct_file*(string in file)
                
                if correct_file:
                    resfilenames.append(folder + "//" + file)
        
                
        if resfilenames != []:
            for file in resfilenames:
            
                f = open(file,'rb')
                res = pickle.load(f)
                f.close()
                
                reslist.append(res)
                
                temp.append(res.temp)
                
                resfreq.append(res.lmfit_vals[1])
                
            indxs = np.argsort(np.array(temp))    
            
            temp = [temp[k] for k in indxs]
            resfreq = [resfreq[k] for k in indxs]
            reslist = [reslist[k] for k in indxs]
    
    
    
    
    
    return temp,resfreq,reslist, resfilenames                
                    

def ExtractLmfitParams(reslist,param = 'f0'):
    
    if param in reslist[0].lmfit_labels:
        indx = reslist[0].lmfit_labels.index(param)
    else:
        print("No %s in the reslist" %(param))
        return [];
    
    params = []
    
    for res in reslist:
        
        params.append(res.lmfit_vals[indx])

    return params


def GetSortIndex(params,desending = True):
    
    indx = sorted(range(len(params)), key=lambda k: params[k])
    
    if desending == False:
        indx.reverse()
    
    return indx
        
def GenColorMap(params,colormap = 'jet'):
    
    color_gen = plt.get_cmap(colormap)
    
    if len(params) > 1:
        
        plt_color = [color_gen((param-min(params))/(max(params)-min(params))) for param in params]
        
        # plt_color = color_gen(params*1.0/max(params))
    else:
        plt_color = [color_gen(0)]
        
    cbar_norm= mpl.colors.Normalize(vmin=min(params), vmax=max(params))

    sm = mpl.cm.ScalarMappable(cmap = colormap,norm = cbar_norm)
    
    return plt_color,sm
    
def GetResTemp(folder,tempindx = 0):
    
    files = os.listdir(folder)
    
    temps = []
    for file in files:
        
        params = ExtractDatafromString(file)
        
        temps.append(params[tempindx])
        
    return temps
        
def SetDefaultPlotParam():
    
    mpl.rcParams['lines.linewidth'] = 2
    # mpl.rcParams['lines.color'] = 'r'    
    mpl.rcParams['axes.labelsize'] = 20;
    
    mpl.rcParams['xtick.labelsize'] = 15;
    mpl.rcParams['ytick.labelsize'] = 15;
    
    mpl.rcParams['font.family'] = 'Times New Roman'
    mpl.rcParams['mathtext.fontset'] = 'custom'
    mpl.rcParams['mathtext.rm'] = 'Times New Roman'
    mpl.rcParams['mathtext.it'] = 'Times New Roman:italic'
    mpl.rcParams['mathtext.bf'] = 'Times New Roman:bold'
    # mpl.rcParams.update({
    # "text.usetex": True,
    # "font.family": "time new roman"})


def GetPulseMaxMin(reslist,pulsewidths = [50]):
    
    pulsemins = []
    pulsemaxs = []
    
    pulsemins_indx = []
    pulsemaxs_indx = []
    
    for pulsewidth in pulsewidths:
        
        indx = reslist[0].pulseWidths.index(pulsewidth)
        
        pulsemin = []
        pulsemax = []
        
        pulseminindx = []
        pulsemaxindx = []
        
        for res in reslist:
            
            pulsephase = res.pulsePhases[indx]
            phase0 = np.mean(np.unwrap(pulsephase[0:10000]))
            
            dphase = np.unwrap(pulsephase - phase0);
            
            pulsemin.append(np.min(dphase))
            pulseminindx.append(np.argmin(dphase))
            
            pulsemax.append(np.max(dphase))
            pulsemaxindx.append(np.argmax(dphase))
            
        pulsemins.append(pulsemin)
        pulsemaxs.append(pulsemax)
        
        pulsemins_indx.append(pulseminindx)
        pulsemaxs_indx.append(pulsemaxindx)
        
    return pulsemins,pulsemaxs,pulsemins_indx,pulsemaxs_indx
    
def CalPowerResponse(oresponse,opower):

    oresponse = np.array(oresponse)
    opower = np.array(opower)
    
    response = []
    
    for i in range(oresponse.shape[1]):
        
        oresponse_i = oresponse[:,i]
        
        fitresult = np.polyfit(opower,oresponse_i,1)
        
        response.append(fitresult[0])
        
        
    return response
    
def GenAveragePulse(I,Q,dt = 1e-8,pulsefreq = 250,t_offset = 0,segtime = 0):
    
    
    NumPointPulse = round(1/pulsefreq/dt);
    
    # pulseI = np.zeros(NumPointPulse);
    # pulseQ = np.zeros(NumPointPulse);
    
    Pointoffset = round(t_offset/dt)
    
    pulseNum = int((len(I)-Pointoffset)/NumPointPulse)
    
    pulseIRaws = []
    pulseQRaws = []
    
    if segtime > 0 and segtime<1/pulsefreq:
        segpoints = round(segtime/dt)
        
    else:
        segpoints = NumPointPulse
        
    pulseI = np.zeros(segpoints);
    pulseQ = np.zeros(segpoints);    
        
    
    for i in range(pulseNum):
        
        I_current = I[(i*NumPointPulse+Pointoffset):(i*NumPointPulse+Pointoffset+segpoints)]
        Q_current = Q[(i*NumPointPulse+Pointoffset):(i*NumPointPulse+Pointoffset+segpoints)]
        pulseIRaws.append(I_current)
        pulseQRaws.append(Q_current)
        
        pulseI = pulseI + I_current
        pulseQ = pulseQ + Q_current
        
        # currentI = 
        
        
    pulseI = pulseI/pulseNum;
    pulseQ = pulseQ/pulseNum;
    
    return pulseI, pulseQ, pulseIRaws,pulseQRaws

# def GenAveragePulse(I,Q,dt = 1e-8,pulsefreq = 250):
    
#     NumPointPulse = round(1/pulsefreq/dt);
    
#     pulseI = np.zeros(NumPointPulse);
#     pulseQ = np.zeros(NumPointPulse);
    
    
#     pulseNum = round(len(I)/NumPointPulse)
    
#     for i in range(pulseNum):
        
#         pulseI = pulseI + I[i*NumPointPulse:(i+1)*NumPointPulse]
#         pulseQ = pulseQ + Q[i*NumPointPulse:(i+1)*NumPointPulse]
        
#     pulseI = pulseI/pulseNum;
#     pulseQ = pulseQ/pulseNum;
    
#     return pulseI, pulseQ
       
    
    


import h5py
import sys
import numpy as np
import pandas as pd
from Functions import Modelisation as mod
import os
import argparse
from pathlib import Path
import pathlib


def parse():
    # read in command line arguments
    parser = argparse.ArgumentParser(description='MKID Pipeline CLI')
    parser.add_argument('--init', action='store_true', help='launch the interface')
    # parser.add_argument('--dinit', action='store_true', help='launch the interface through docker')
    parser.add_argument('--outp', help='output destination', dest='out_cfg', default=None)
    parser.add_argument('--inp', help='data location', dest='in_cfg' , default=None)
    parser.add_argument('--format', help='format of the plot', default='.jpg', dest='form')
    parser.add_argument('--dir', help='create input and output folder', action='store_true')
    return parser.parse_args()


if __name__ == "__main__":
    args = parse()

    if args.init:
        try:
            path = str(sorted(pathlib.Path('/').glob('**/Interface.py'))[0]) #Docker version
        except:
            path = str(sorted(pathlib.Path('').glob('**/Interface.py'))[0])  #Terminal version
        os.system("streamlit run "+path)
    
    if args.out_cfg!=None and args.in_cfg!=None:
        mod.Data_read(args.in_cfg,args.out_cfg,args.form)
        print('Done')

    if args.dir:
        output = Path('Output')
        Input = Path('Data_location')
        if output.exists():
            pass
        else:
            output.mkdir()
        if Input.exists():
            pass
        else:
            Input.mkdir()



from tkinter import * 
from tkinter.filedialog import *
from tkinter import ttk
import Functions.Modelisation as Model
import matplotlib.pyplot as plt



def interface():


    def path_data():
        filepath = askopenfilename(title="Path",filetypes=[('hdf files','.hdf'),('all files','.*')])
        value_data.set(filepath)


    def path_plot():
        filepath = askdirectory()
        value_plot.set(filepath)


    def Launch():
        Model.Data_read(value_data.get(),value_plot.get(),FormatMenu.get())
        Main_Window.quit()

 

    Main_Window = Tk()
    Main_Window.title('DRS')
    frm = ttk.Frame(Main_Window, padding=10)
    frm.grid()

    value_data = StringVar() 
    FormatList = ['.jpg','.svg','.eps']

    # Taking the data location, HDF5 required
    label = ttk.Label(frm,text="Data location ?").grid(column = 0,row=0)
    entry_data = ttk.Entry(frm, width=30,textvariable=value_data).grid(column=0,row=1)
    button_data = ttk.Button(frm,text='Browse',command=path_data).grid(column=1,row=1)

    label = ttk.Label(frm,text = '                              ').grid(column=2,row=0)

    # Taking Folder and Format for the plot saving
    label = ttk.Label(frm, text = 'Format ?').grid(column=5,row=0)
    label=ttk.Label(frm,text="Results").grid(column=3,row=0)
    button_data = ttk.Button(frm,text='Browse',command=path_plot).grid(column=4,row=1)
    value_plot = StringVar() 
    entry_plot = ttk.Entry(frm, width=30,textvariable=value_plot).grid(column=3,row=1)
    FormatMenu = ttk.Combobox(frm,values=FormatList)
    FormatMenu.current(0)
    FormatMenu.grid(column=5,row=1)

    #Launching the code
    Launch_button=ttk.Button(frm,text="OK", command=Launch).grid(column=0,row=3)
    


    Main_Window.mainloop()



import streamlit as st
import pandas as pd
import numpy as np
import Functions.Modelisation as Model

st.title('SPIAKID Data Reduction System')


col1, col2, col3, col4 = st.columns([1, 0.5, 1, 1])

with col3:
    result_folder = st.text_input('Result folder',value="")

with col4:
    format = st.selectbox('Format', ['.jpg','.svg','.eps'])
with col1:
    data_location = st.text_input('Data Location',value="")
    Launch = st.button('OK')
    if Launch:
        Model.Data_read(data_location,result_folder,format)





import numpy as np
import h5py
import matplotlib.pyplot as plt
import lmfit as lf
import glob
import scipy.signal as sps
from scipy import interpolate as interp

import Functions.resonator as a
import Functions.SinglePhotonProcess as spp
import Functions.cmplxIQ as fitmodel

def Data_read(Data,Plot_path,Plot_Format):
    f = h5py.File(Data,'r')
#get the temp and readout power of the measurement
    temp = f['header'].attrs['temp']
    pwr = f['header'].attrs['power']

#get frequency, I and Q of the resonator 
    freq = f['resonator/freq'][...]
    I = f['resonator/I'][...]
    Q = f['resonator/Q'][...]

    I0 = f['resonator/I0'][...]
    Q0 = f['resonator/Q0'][...]

    #get the IQ calibation data for the noise and pulse data
    IQCaldata = f['calibration/IQCaldata'][...]

    #create a resonator object
    res = a.Resonator('1', temp, pwr, freq, I, Q,I0 = I0,Q0 = Q0)
    #fit the resonator with S21 (transmission)
    res.load_params(fitmodel.cmplxIQ_params)
    res.do_lmfit(fitmodel.cmplxIQ_fit)

    t_offset = 0.001
    binsize = 500

    templatetime = 0.002

    laserfreq = f['pulse'].attrs['laserfreq']
    pulsetriggertime = 1/laserfreq/2 - t_offset
    savefolder = Plot_path
    samplefreq = f['pulse'].attrs['samplefreq']

    #generate the averaged pulse data as the template
    avedata = spp.GenPulseAverageWithHDF(res,f,savefolder,IQCaldata,calibrate_drift = True, t_offset = t_offset,savefig = True,Format=Plot_Format)

    #generate the psd (Power Spectral Density) of the resonator
    psd = spp.GenPSDWithHDF(res,f,savefolder,IQCaldata,binsize = binsize,pulselaser = laserfreq,Format=Plot_Format)


    #generate the wiener filter for both amplitude and phase
    wifilterphase, wifilteramp,indx = spp.GenWienerFilter(avedata,psd,savefolder,binsize = binsize,pulsetriggertime = pulsetriggertime,dt = 1/samplefreq, templatetime = templatetime,Format=Plot_Format)

    #use the wiener filter to generate the pulse statistics
    pulseheights = spp.GenPulseStatisticsHDF(res,f,wifilterphase,savefolder,IQCaldata,binsize = binsize,t_offset = t_offset,calibrate_drift = True,pulsetype = 'phase',plotbin = 0.2,Format=Plot_Format)

    #estimate the energy resolution
    spp.EnergyResolution(pulseheights,savefolder,photonum = 6,Format=Plot_Format)






import numpy as np
import matplotlib.pyplot as plt
import pickle
import Functions.utility as ut
import scipy.signal as sg
from lmfit.models import GaussianModel
import Functions.IQCalibration as IQCal

def GenPulseAverage(res,pulsefolder,savefolder,IQCaldata,pulselaser = 250,
                    dt = 1e-8, NumpulseUnscaled = 80,
                    calibrate_drift = True, t_offset = 0.001,savefig = True):
    
    
    pulsePointNum = round(1/pulselaser/dt)
    Ip_av = np.zeros(pulsePointNum)
    Qp_av = np.zeros(pulsePointNum)
    
    processedpulse = 0

    # Idrifts = []
    # Qdrifts = []

    # calibrate_drift = True
    for i in range(NumpulseUnscaled):
        
        # if i not in usfulseg:
            # continue 
        processedpulse += 1;
        filename = pulsefolder + '//%d.obj' %(i)
        
        f = open(filename,'rb')
        pulsedata = pickle.load(f)
        f.close()
        
        print("i:",i)
        # pulsedata = res.pulseUnscaled[i]
        I0 = res.I0;
        Q0 = res.Q0;
        freq = res.freq;
        
        I,Q,pulsemeasfreq,fsample = ut.CalPulseRaw(pulsedata,IQCaldata,freq,I0,Q0,ispulse = True)
        # t = np.arange(0,len(I))*1e-08
        
        # If,Qf = ut.lowpassfiltering(I,100e6,Q,cutoff = 250e3)
        resfreqindx = np.argmin(np.abs(res.freq-pulsemeasfreq*1e9))
        
        
        if i==0:

            Ip_initial = res.I[resfreqindx]
            Qp_initial = res.Q[resfreqindx]
    
        Ip,Qp,pulseIRaws,pulseQRaws = ut.GenAveragePulse(I,Q,dt = dt,pulsefreq = pulselaser,t_offset = t_offset)
        
        points = int((1/pulselaser/2 - t_offset)/dt)
        
        # Idrifts.append(np.mean(Ip[0:10000]))
        # Qdrifts.append(np.mean(Qp[0:10000]))
        # print(np.mean(Ip[0:10000]))
        # print(np.mean(Qp[0:10000]))
        
        if calibrate_drift:
            Ip = Ip - np.mean(Ip[0:points]) + Ip_initial 
            Qp = Qp - np.mean(Qp[0:points]) + Qp_initial 
        
        
        
        Ip_av = Ip + Ip_av;
        Qp_av = Qp + Qp_av;
        
        
    t_seg = np.arange(0,len(Ip_av))*dt
        

    pulsefreq = pulsedata['freq']

    res.pulseUnscaled = []



    Ip_av = Ip_av/processedpulse;
    Qp_av = Qp_av/processedpulse;

    amp,phase = res.cal_pulse_outside(Ip_av + 1j*Qp_av,pulsedata['freq']*1e9,pulsewidth = 50,fs = 200e3)

    data = {"pulsefreq":pulsefreq,
            'res':res,
            't':t_seg,
            'Ia':Ip_av,
            'Qa':Qp_av,
            'amp':amp,
            'phase':phase,
            't_offset':t_offset,
            'dt':dt}

    f = open(savefolder + '/averagedpulse.obj','wb')
    pickle.dump(data,f)

    f.close()
    
    if savefig == True:

        plt.figure()
        
        points = int((1/pulselaser/2 - t_offset)/dt)
        phase0 = np.mean(phase[0:points])
        amp0 = np.mean(amp[0:points])

        phasedeg = -(phase-phase0);

        phasedeg = np.unwrap(phasedeg)
        damp = -(amp-amp0)
        plt.plot(t_seg*1000-t_offset*1000,phasedeg/np.max(phasedeg),label = 'phase')
        plt.plot(t_seg*1000-t_offset*1000,damp/np.max(phasedeg),label = 'amp')


        plt.xlabel('Time(ms)')
        plt.ylabel('Normalized response')

        plt.legend(fontsize = 15)
        plt.grid()

        plt.savefig(savefolder + '/average pulse phase response-norm.svg',bbox_inches = 'tight')
        plt.savefig(savefolder + '/average pulse phase response-norm.jpg',dpi = 600,bbox_inches = 'tight')
        plt.savefig(savefolder + '/average pulse phase response-norm.eps',bbox_inches = 'tight')
        
        # plt.close()
        plt.figure()

        plt.plot(t_seg*1000-t_offset*1000,phasedeg,label = 'phase')
        plt.plot(t_seg*1000-t_offset*1000,damp,label = 'amp')

        plt.xlabel('Time(ms)')
        plt.ylabel('Response(rad)')

        plt.legend(fontsize = 15)
        plt.grid()

        plt.savefig(savefolder + '/average pulse phase response.svg',bbox_inches = 'tight')
        plt.savefig(savefolder + '/average pulse phase response.jpg',dpi = 600,bbox_inches = 'tight')
        plt.savefig(savefolder + '/average pulse phase response.eps',bbox_inches = 'tight')
        # plt.close()
        
    return data


def GenPulseAverageWithHDF(res,hdfile,savefolder,IQCaldata,
                    calibrate_drift = True, t_offset = 0.001,savefig = True,Format = '.jpg'):
    
    #res: scraps.resonator obj
    
    #hdfile: the hdfile contain the pulse data
    
    #savefolder: directory to save the processed the data
    
    #IQCaldata: IQ Calibration data for the raw pulse data
    
    #calibrate_drift: bool, to remove the drift in the baseline or not. 
    
    #t_offset: to move the trigger time of the pulse. t_offset = 0 means the trigger of the pulse is in the center
        
    #savefig: bool, to produce the plot of the averaged data or not
    
    #load the dataset of the pulse in the hdfile
    pulse = hdfile['pulse']
    
    #the frequency of the trigger of the laser.
    pulselaser = pulse.attrs['laserfreq']
    
    #the data of the pulse is segmented. Load the number of the groups
    NumpulseUnscaled = pulse.attrs['groupnum']
    print(NumpulseUnscaled)
    #load the sample frequency of the data. The usual sample rate is about 100MHz
    dt = 1/pulse.attrs['samplefreq']
    
    #calculate the number of data point in one period of the laser pulse
    pulsePointNum = round(1/pulselaser/dt)
    Ip_av = np.zeros(pulsePointNum)
    Qp_av = np.zeros(pulsePointNum)

    # calibrate_drift = True
    for i in range(NumpulseUnscaled):
        
        #load the raw data of the IQ data, which is numpy.int16 
        pulsedata = pulse['IQ%d'%(i)][...]
        
        #load the voltage gain and the offset of the oscilloscope 
        vgains = pulse['vgains'][...]
        voffsets = pulse['voffsets'][...]

        print("i:",i)
        
        #load the DC offset of the IQ from the res
        I0 = res.I0;
        Q0 = res.Q0;
        freq = res.freq;
        
        #load the readout tone frequency
        pulsefreq = pulse.attrs['measfreq']
        
        #scale the data from the oscillscope
        I = pulsedata[0]*vgains[0] - voffsets[0]
        Q = pulsedata[1]*vgains[1] - voffsets[1]
        
        #calibrate the IQ imbalance from the IQ mixer, i.e. the DC offset, amplitude and phase imbalance
        I,Q = IQCal.IQ_CalNoise(I,Q,pulsefreq/1e9,freq/1e9,I0,Q0,IQCaldata)
        
        resfreqindx = np.argmin(np.abs(res.freq-pulsefreq))
        
        #produce the reference for the I and Q of the pulse
        if i==0:

            Ip_initial = res.I[resfreqindx]
            Qp_initial = res.Q[resfreqindx]
        
        #segment the data by laser frequency
        Ip,Qp,pulseIRaws,pulseQRaws = ut.GenAveragePulse(I,Q,dt = dt,pulsefreq = pulselaser,t_offset = t_offset)
        
        #to remove the DC offset or not
        points = int((1/pulselaser/2 - t_offset)/dt)
        if calibrate_drift:
            Ip = Ip - np.mean(Ip[0:points]) + Ip_initial 
            Qp = Qp - np.mean(Qp[0:points]) + Qp_initial 
        
        #calculate the average of the pulse
        Ip_av = Ip + Ip_av;
        Qp_av = Qp + Qp_av;
        
    
    t_seg = np.arange(0,len(Ip_av))*dt
        
    # res.pulseUnscaled = []

    #generate the IQ of the pulse
    Ip_av = Ip_av/NumpulseUnscaled;
    Qp_av = Qp_av/NumpulseUnscaled;
    
    #calculate the amplitude and phase response of the MKIDs
    amp,phase = res.cal_pulse_outside(Ip_av + 1j*Qp_av,pulsefreq,pulsewidth = 50)

    data = {"pulsefreq":pulsefreq,
            'res':res,
            't':t_seg,
            'Ia':Ip_av,
            'Qa':Qp_av,
            'amp':amp,
            'phase':phase,
            't_offset':t_offset,
            'dt':dt}

    #f = open(savefolder + '/averagedpulse.obj','wb')
    #pickle.dump(data,f)
    #f.close()
    
    #produce the plot of the averaged pulse and amplitude of the detector
    if savefig == True:

        plt.figure(1)
        plt.title('Avedata')
        plt.subplot(1,2,1)
        points = int((1/pulselaser/2 - t_offset)/dt)
        phase0 = np.mean(phase[0:points])
        amp0 = np.mean(amp[0:points])

        phasedeg = -(phase-phase0);

        phasedeg = np.unwrap(phasedeg)
        damp = -(amp-amp0)
        plt.plot(t_seg*1000-t_offset*1000,phasedeg/np.max(phasedeg),label = 'phase')
        plt.plot(t_seg*1000-t_offset*1000,damp/np.max(phasedeg),label = 'amp')


        plt.xlabel('Time(ms)')
        plt.ylabel('Normalized response')

        plt.legend(fontsize = 15)
        plt.grid()

        # plt.savefig(savefolder + '/average pulse phase response-norm.svg',bbox_inches = 'tight')
        # plt.savefig(savefolder + '/average pulse phase response-norm.jpg',dpi = 600,bbox_inches = 'tight')
        # plt.savefig(savefolder + '/average pulse phase response-norm.eps',bbox_inches = 'tight')
        
        # plt.close()
      
        plt.subplot(1,2,2)
        plt.plot(t_seg*1000-t_offset*1000,phasedeg,label = 'phase')
        plt.plot(t_seg*1000-t_offset*1000,damp,label = 'amp')

        plt.xlabel('Time(ms)')
        plt.ylabel('Response(rad)')

        plt.legend(fontsize = 15)
        plt.grid()
        
        plt.savefig(savefolder + '/avedata' + Format ,bbox_inches = 'tight')
   
        # plt.savefig(savefolder + '/average pulse phase response.jpg',dpi = 600,bbox_inches = 'tight')
        # plt.savefig(savefolder + '/average pulse phase response.eps',bbox_inches = 'tight')
        # plt.close()
        
    return data

# hdfile

def GenPSDWithHDF(res,hdfile,savefolder,IQCaldata,binsize = 100,
           pulselaser = 250, NoiseSeglength = None, plotpsd = True,Format = ".jpg"):
    
    #res: scraps.resonator obj
    
    #hdfile: the hdfile contain the pulse data
    
    #savefolder: directory to save the processed the data
    
    #IQCaldata: IQ Calibration data for the raw pulse data
    
    #binsize: the data point of averaging the data
      
    #plotpsd: bool, to produce the plot of the averaged data or not
    
    #NoiseSeglength: the length of the data points to perform fft
    
    In,Qn= ut.CalNoiseHDf(res,hdfile,IQCaldata)
    
    fsample = hdfile['noise'].attrs['samplefreq']
    noisefreq = hdfile['noise'].attrs['measfreq']
    
    pulsePointNum = round(1/pulselaser*fsample);
    
    
    if  NoiseSeglength == None:
        NoiseSeglength = int(pulsePointNum/binsize)

    # In = np.concatenate(noiseIs)
    # Qn = np.concatenate(noiseQs)
    
    t_seg = np.arange(0,len(In))*1/fsample
    
    t_bin,In_bined = ut.AverageArray(t_seg,In,chuncksize = binsize)
    t_bin,Qn_bined = ut.AverageArray(t_seg,Qn,chuncksize = binsize)
    
    #calculate the amplitude and phase of the noise spectrum
    pulse_amp,pulse_phase = res.cal_pulse_outside(In_bined + 1j*Qn_bined,noisefreq,pulsewidth = 50,fs = 1/(t_bin[1]-t_bin[0]))
    #remove the 2*pi change
    pulse_phase = np.unwrap(pulse_phase)
    
    #calculate the noise spectrum with welth method in the amplitude and phase direction. 
    fpsd,pxx = sg.welch(pulse_phase-np.mean(pulse_phase),fs = 1/(t_bin[1]-t_bin[0]),nperseg = NoiseSeglength)
    fpsd,pyy = sg.welch(pulse_amp-np.mean(pulse_amp),fs = 1/(t_bin[1]-t_bin[0]),nperseg = NoiseSeglength)
    data = np.vstack([fpsd,pxx,pyy])
    
    np.savetxt(savefolder + '// psd%d.txt'%(binsize),data.T)
    
    #plot the psd
    if plotpsd == True:
        
        ut.SetDefaultPlotParam()
        plt.figure()
        plt.semilogx(fpsd,pxx,label='phase')
        plt.semilogx(fpsd,pyy,label='amp')
        plt.xlabel("Freq(Hz)")
        plt.ylabel("Psd($rad^2/Hz)$")
        plt.legend()
        plt.savefig(savefolder + "//psd%d.jpg"%(binsize),dpi = 600,bbox_inches = 'tight')
        
    return data.T

def GenPSDFromNoise(res,IQCaldata,savefolder,binsize = 100,
           pulselaser = 250,plotpsd = True):
    
    In,Qn,noisefreq,fsample = ut.CalNoiseRaw(res,IQCaldata,isonres = True)
    
    pulsePointNum = round(1/pulselaser*fsample);
    NoiseSeglength = int(pulsePointNum/binsize)

    # In = np.concatenate(noiseIs)
    # Qn = np.concatenate(noiseQs)
    
    t_seg = np.arange(0,len(In))*1/fsample
    
    t_bin,In_bined = ut.AverageArray(t_seg,In,chuncksize = binsize)
    t_bin,Qn_bined = ut.AverageArray(t_seg,Qn,chuncksize = binsize)
    
    #calculate the pulse
    pulse_amp,pulse_phase = res.cal_pulse_outside(In_bined + 1j*Qn_bined,noisefreq*1e9,pulsewidth = 50,fs = 1/(t_bin[1]-t_bin[0]))
    #remove the 2*pi change
    pulse_phase = np.unwrap(pulse_phase)
    
    fpsd,pxx = sg.welch(pulse_phase-np.mean(pulse_phase),fs = 1/(t_bin[1]-t_bin[0]),nperseg = NoiseSeglength)
    fpsd,pyy = sg.welch(pulse_amp-np.mean(pulse_amp),fs = 1/(t_bin[1]-t_bin[0]),nperseg = NoiseSeglength)
    data = np.vstack([fpsd,pxx,pyy])
    
    np.savetxt(savefolder + '// psd%d.txt'%(binsize),data.T)
    
    if plotpsd == True:
        
        ut.SetDefaultPlotParam()
        plt.figure()
        plt.semilogx(fpsd,pxx)
        plt.semilogx(fpsd,pyy)
        plt.xlabel("Freq(Hz)")
        plt.ylabel("Psd($rad^2/Hz$")
        
        plt.savefig(savefolder + "//psd%d.jpg"%(binsize),dpi = 600,bbox_inches = 'tight')
    
    return data.T


def GenPSD(res,pulsefolder,IQCaldata,savefolder,binsize = 100,pulseNum = 10,
           pulselaser = 250,dt = 1e-8,t_offset = 0.001,calibrate_drift = False):
    
    pulsePointNum = round(1/pulselaser/dt);
    
    NoiseSeglength = int(pulsePointNum/binsize)
    Ip_av = np.zeros(pulsePointNum);
    # Qp_av = np.zeros(pulsePointNum)
    
    # f = open(IQfile,'rb')
    # res = pickle.load(f)
    # f.close()
    
    pxx_all = []
    pyy_all = []
    
    for i in range(pulseNum):
        
        print("Current i:",i)
        # pulsedata = res.pulseUnscaled[i]
        f = open(pulsefolder + '/%d.obj' %(i),'rb')
        pulsedata = pickle.load(f)
        f.close()
        
        I0 = res.I0;
        Q0 = res.Q0;
        freq = res.freq;
        

        I,Q,measfreq,fsample = ut.CalPulseRaw(pulsedata,IQCaldata,freq,I0,Q0,ispulse = True)
        # t = np.arange(0,len(I))*1e-08

        Ip,Qp,pulseIRaws,pulseQRaws = ut.GenAveragePulse(I,Q,dt = dt,
                                                         pulsefreq = pulselaser,
                                                         t_offset = t_offset)
        
        resfreqindx = np.argmin(np.abs(res.freq-measfreq*1e9))
        
        if i==0:

            Ip_initial = res.I[resfreqindx]
            Qp_initial = res.Q[resfreqindx]
        
        
        # Ip_av = Ip + Ip_av;
        # Qp_av = Qp + Qp_av;
        
        t_seg = np.arange(0,len(Ip_av))*dt
        
        I_noise_all = []
        Q_noise_all = []
        
    
        for pulseindx in range(len(pulseIRaws)):
            
            Iv = pulseIRaws[pulseindx]
            Qv = pulseQRaws[pulseindx]
            
            points = int((1/pulselaser/2-t_offset)/dt*3/4);
            
            Iv_noise = Iv[0:points]
            Qv_noise = Qv[0:points]
            
            t_bin,Iv_bined = ut.AverageArray(t_seg,Iv_noise,chuncksize = binsize)
            t_bin,Qv_bined = ut.AverageArray(t_seg,Qv_noise,chuncksize = binsize)
            
            
            #calibrate the offset, it assumes that the offset mainly comes from the amplifiers
            #it could also comes from the fact that the temperature of the MKIDs is rising. In this
            #case, it would be difficult to calibrate the data. 
            
            if calibrate_drift:
                Iv_bined = Iv_bined - np.mean(Iv_bined) + Ip_initial 
                Qv_bined = Qv_bined - np.mean(Qv_bined) + Qp_initial 
            
            
            I_noise_all.append(Iv_bined)
            Q_noise_all.append(Qv_bined)
        
        I_noise_all = np.concatenate(I_noise_all)
        Q_noise_all = np.concatenate(Q_noise_all)
        
        pulse_amp,pulse_phase = res.cal_pulse_outside(I_noise_all + 1j*Q_noise_all,pulsedata['freq']*1e9,pulsewidth = 50,fs = 1/(t_bin[1]-t_bin[0]))

        pulse_phase = np.unwrap(pulse_phase)
        
        fpsd,pxx = sg.welch(pulse_phase-np.mean(pulse_phase),fs = 1/(t_bin[1]-t_bin[0]),nperseg = NoiseSeglength)
        
        fpsd,pyy = sg.welch(pulse_amp-np.mean(pulse_amp),fs = 1/(t_bin[1]-t_bin[0]),nperseg = NoiseSeglength)
            
        pxx_all.append(pxx)
        pyy_all.append(pyy)
    
    pxx_average = np.zeros(np.shape(pxx))
    pyy_average = np.zeros(np.shape(pyy))


    for pxx,pyy in zip(pxx_all,pyy_all):
        
        pxx_average = pxx_average + pxx;
        pyy_average = pyy_average + pyy;

    pxx_average = pxx_average/len(pxx_all)
    pyy_average = pyy_average/len(pyy_all)
        
    
    data = np.vstack([fpsd,pxx_average,pyy_average])
    
    np.savetxt(savefolder + '//psd%d.txt'%(binsize),data.T)
    
    plt.figure()
    plt.loglog(fpsd[1:-1],pxx_average[1:-1],label = 'Phase')
    plt.loglog(fpsd[1:-1],pyy_average[1:-1],label = 'Amp')
    plt.xlabel('Freq(Hz)')
    plt.ylabel('PSD($rad^2/Hz$)')
    plt.title("dt = %fus"%(binsize*dt*1e6))
    plt.savefig(savefolder + '//psd%d.jpg' %(binsize),dpi = 300,bbox_inches = 'tight')
    # plt.close()
    
    return data.T

def GenWienerFilter(templatedata,psd,savefolder,binsize = 500,pulsetriggertime = 0.001,
                    dt = 1e-8, templatetime = 0.003,
                    templatetype = 'average', pulsedirection = 'negtive',Format = ".jpg"):
    
    # data = np.loadtxt("data 3.8885GHz/20220629133637-3.888GHz/Res Index 0/template_average_500.txt")

    # psd = np.loadtxt('20220709164518-4.9 sccm - batch 2/Res Index 1/Temp 50.0mK/noise psd bin 500.txt')
    if templatetype == 'average':
        # f = open(templatefile,'rb')
        # pulsesinglephoton = pickle.load(f)
        # f.close()

        PI = templatedata['Ia']
        PQ = templatedata['Qa']
        t_seg = templatedata['t']
        
    else: #the template type is single photon pulse

        pass
    

    t_bin,Iv_bined = ut.AverageArray(t_seg,PI,chuncksize = binsize)
    t_bin,Qv_bined = ut.AverageArray(t_seg,PQ,chuncksize = binsize)

    res = templatedata['res']

    # f.close()
    pulsefreq = (res.lmfit_vals[1] + res.lmfit_vals[0])
    amp,phase = res.cal_pulse_outside(Iv_bined + 1j*Qv_bined,pulsefreq)

    trigerindx = int(pulsetriggertime/dt/binsize)
    
    # print(trigerindx)

    phase0 = np.mean(phase[0:trigerindx])
    phase1 = phase - phase0
    phase1 = np.unwrap(phase1)
    
    amp0 = np.mean(amp[0:trigerindx])
    amp1 = amp - amp0

        
    if pulsedirection == 'positive':
        templatephase = phase1/np.max(phase1)
        
    else:
        templatephase = -phase1/np.min(phase1)
        templateamp = -amp1/np.min(amp1)
    
    templatelength = int(templatetime/dt/binsize)
    
    psdlength = (len(psd)-1)*2
    
    wifilterphase = ut.wiener(templatephase[0:templatelength],psd[:,1],psdlength)
    wifilteramp = ut.wiener(templatephase[0:templatelength],psd[:,2],psdlength)
    # wifilter = wiener(templatephase,psd[:,1],800)

    filtered_templatephase = sg.lfilter(wifilterphase,1,templatephase)
    filtered_templateamp = sg.lfilter(wifilteramp,1,templateamp)
    
    indx = np.min(filtered_templatephase)

    plt.figure()
    plt.plot(filtered_templatephase,label = 'filtered_template')
    plt.plot(templatephase[0:templatelength],label = 'template phase')
    plt.plot(wifilterphase,label = 'wiener filter phase');
    
    plt.xlabel("Index")
    plt.ylabel('Pulse Height')
    plt.legend(fontsize = 18)
    plt.savefig(savefolder + '//filtered template - phase'+Format,bbox_inches = 'tight')
    # plt.close()
    
    plt.figure()
    plt.plot(filtered_templateamp,label = 'filtered_template')
    plt.plot(templateamp[0:templatelength],label = 'template amp')
    plt.plot(wifilteramp,label = 'wiener filter amp');
    
    plt.xlabel("Index")
    plt.ylabel('Pulse Height')
    plt.legend(fontsize = 18)
    plt.savefig(savefolder + '//filtered template - amp'+Format,bbox_inches = 'tight')
    # plt.close()
    
    data = np.vstack([wifilterphase,wifilteramp])
    
    np.savetxt(savefolder + '/wienner filter-average%d.txt' %(binsize),data.T)


    return wifilterphase, wifilteramp,indx

def GenPulseStatisticsHDF(res,hdfile,wienerfilter,savefolder,IQCaldata,
                       binsize = 500,t_offset = 0.001,calibrate_drift = True,
                       pulsetype = 'phase',lowpassfiltering = False,filterorder = 10,plotbin = 0.2,Format = '.jpg'):
    
    #res: scraps.resonator obj
    
    #hdfile: the hdfile contain the pulse data
    
    #wienerfilter: the wiener filter to process the data
    
    #savefolder: directory to save the processed the data
    
    #IQCaldata: IQ Calibration data for the raw pulse data
    
    #binsize: int, the average data number
    
    #t_offset: to move the trigger time of the pulse. t_offset = 0 means the trigger of the pulse is in the center
    
    #calibrate_drift: bool, to remove the drift in the baseline or not. 
    
    #pulsetype: 'phase', or 'amp', to make the statisics using phase or amplitude
    
    #lowpassfiltering: bool, to do the lowpass filtering of the pulse data
    
    #filterorder: int, the order of the lowpass filter.
    
    #plotbin: the bin size for plotting the statistics
       
    # savefig: bool, to produce the plot of the averaged data or not
    
    #load the dataset of the pulse in the hdfile
    pulse = hdfile['pulse']
    
    #the frequency of the trigger of the laser.
    pulselaser = pulse.attrs['laserfreq']
    
    #the data of the pulse is segmented. Load the number of the groups
    NumpulseUnscaled = pulse.attrs['groupnum']
    
    #load the sample frequency of the data. The usual sample rate is about 100MHz
    dt = 1/pulse.attrs['samplefreq']

    pulse_height = []
    pulse_height_2 = []
    
    #load the DC offset of the IQ from the res
    I0 = res.I0;
    Q0 = res.Q0;
    freq = res.freq;
    
    #load the voltage gain and the offset of the oscilloscope 
    vgains = pulse['vgains'][...]
    voffsets = pulse['voffsets'][...]

    #load the readout tone frequency
    measfreq = pulse.attrs['measfreq']

    for i in range(NumpulseUnscaled):
        
        print("i:",i)
        
        #load the raw data of the IQ data, which is numpy.int16 
        pulsedata = pulse['IQ%d'%(i)][...]
        
        #scale the data from the oscillscope
        I = pulsedata[0]*vgains[0] - voffsets[0]
        Q = pulsedata[1]*vgains[1] - voffsets[1]
        
        #calibrate the IQ imbalance from the IQ mixer, i.e. the DC offset, amplitude and phase imbalance
        I,Q = IQCal.IQ_CalNoise(I,Q,measfreq/1e9,freq/1e9,I0,Q0,IQCaldata)
        
        resfreqindx = np.argmin(np.abs(res.freq-measfreq))
        
        #produce the reference for the I and Q of the pulse
        if i==0:

            Ip_initial = res.I[resfreqindx]
            Qp_initial = res.Q[resfreqindx]
        
        #segment the data by laser frequency
        Ip,Qp,pulseIRaws,pulseQRaws = ut.GenAveragePulse(I,Q,dt = dt,pulsefreq = pulselaser,t_offset = t_offset)
         
        t_seg = np.arange(0,len(Ip))*dt
        
        for pulseindx in range(len(pulseIRaws)):
            
            Iv = pulseIRaws[pulseindx]
            Qv = pulseQRaws[pulseindx]

            points = int((1/pulselaser/2 - t_offset)/dt)
            
            if calibrate_drift:
                Iv = Iv - np.mean(Iv[0:points]) + Ip_initial
                Qv = Qv - np.mean(Qv[0:points]) + Qp_initial
            
            t_bin,Iv_bined = ut.AverageArray(t_seg,Iv,chuncksize = binsize)
            t_bin,Qv_bined = ut.AverageArray(t_seg,Qv,chuncksize = binsize)
            pulse_amp,pulse_phase = res.cal_pulse_outside(Iv_bined + 1j*Qv_bined,measfreq)
            
            avenum = int((1/pulselaser/2 - t_offset)/binsize/dt)
            
            if pulsetype == 'phase':
                
                pulse_phase = np.unwrap(pulse_phase)

                phase0 = np.mean(pulse_phase[0:avenum])
            
                pulse_degree = (pulse_phase-phase0)
            
                pulse_degree = np.unwrap(pulse_degree)*180/np.pi
                pulse_wiener = sg.lfilter(wienerfilter,1,pulse_degree)
            
                indxpulse = len(wienerfilter)
            
                p_min = np.min(pulse_wiener[indxpulse-5:indxpulse+5])
            
                pulse_height.append(p_min)
            
                pulse_height_2.append(np.min(pulse_wiener[indxpulse-2:indxpulse+2]))
            
            else:
                
                amp0 = np.mean(pulse_amp[0:avenum])
            
                pulse_amp = (pulse_amp-amp0)
            
                pulse_amp = pulse_amp*180/np.pi
                pulse_wiener = sg.lfilter(wienerfilter,1,pulse_amp)
            
                indxpulse = len(wienerfilter)
            
                p_min = np.min(pulse_wiener[indxpulse-5:indxpulse+5])
            
                pulse_height.append(p_min)
            
                pulse_height_2.append(np.min(pulse_wiener[indxpulse-2:indxpulse+2]))
            

    # if pulsetype == 'phase':
    p_array = -np.array(pulse_height)
    pulsemin = np.min(p_array)
    pulsemax = np.max(p_array)

    bins = np.arange(pulsemin,pulsemax,plotbin)
    
    plt.figure()
    plt.hist(p_array,bins = bins)    

    plt.xlabel('Pulse height (deg)',fontsize = 20)
    plt.ylabel('Counts',fontsize = 20)
    
    if pulsetype == 'phase':
        plt.savefig(savefolder + '/pulse statistics-average%d'+Format,bbox_inches = 'tight')
        # plt.savefig(savefolder + '/pulse statistics-average%d.jpg'%(binsize),bbox_inches = 'tight')
        # plt.savefig(savefolder + '/pulse statistics-average%d.svg'%(binsize),bbox_inches = 'tight')
    
    else:
        plt.savefig(savefolder + '/pulse statistics-average-amp%d'+Format,bbox_inches = 'tight')
        # plt.savefig(savefolder + '/pulse statistics-average-amp%d.jpg'%(binsize),bbox_inches = 'tight')
        # plt.savefig(savefolder + '/pulse statistics-average-amp%d.svg'%(binsize),bbox_inches = 'tight')
    
    
    # plt.close()
    
    if pulsetype == 'phase':
        np.savetxt(savefolder + '/pulse_height-average%d.txt'%(binsize),pulse_height)
        np.savetxt(savefolder + '/pulse_height-average%d_2.txt'%(binsize),pulse_height_2)
        
    else:
        np.savetxt(savefolder + '/pulse_height-average-amp%d.txt'%(binsize),pulse_height)
        np.savetxt(savefolder + '/pulse_height-average-amp%d_2.txt'%(binsize),pulse_height_2)

    return pulse_height


def GenPulseStatistics(res,binsize,wienerfilter,IQCaldata,savefolder,
                       pulsefolder,calibrate_offset = True,dt = 1e-8,
                       pulselaser = 250,NumpulseUnscaled=80,t_offset = 0.001,
                       pulsetype = 'phase',lowpassfiltering = False,filterorder = 10):
    

    pulsePointNum = round(1/pulselaser/dt);
    Ip_av = np.zeros(pulsePointNum);
    Qp_av = np.zeros(pulsePointNum)

    pulse_height = []
    pulse_height_2 = []
    
    # pulse_height_amp = []
    # pulse_height_amp_2 = []

    pulseprocess = NumpulseUnscaled;


    for i in range(pulseprocess):
        
        print("i:",i)

        f = open(pulsefolder + '/%d.obj'%(i),'rb')
        pulsedata = pickle.load(f)
        f.close()
        
        
        I0 = res.I0;
        Q0 = res.Q0;
        freq = res.freq;

        I,Q,pulsefreq,fsample = ut.CalPulseRaw(pulsedata,IQCaldata,freq,I0,Q0,ispulse = True)
        # t = np.arange(0,len(I))*1e-08
        
        if lowpassfiltering == True:
            I,Q = ut.lowpassfiltering(I, sample_freq = 1/dt,Q = Q,cutoff = 100e3,filter_order = 10)
        
        Ip,Qp,pulseIRaws,pulseQRaws = ut.GenAveragePulse(I,Q,dt = dt,
                                                         pulsefreq = pulselaser,t_offset = t_offset)
        
        resfreqindx = np.argmin(np.abs(res.freq-pulsefreq*1e9))
        
        
        if i==0:

            Ip_initial = res.I[resfreqindx]
            Qp_initial = res.Q[resfreqindx]
        
        #take the data of 2ms before the pulse for the noise calculation 
        # Ip1,Qp1,noiseIRaws,noiseQRaws = ut.GenAveragePulse(I,Q,dt = 1e-8,pulsefreq = pulselaser,
                                                           # t_offset = 0.000,segtime = 0.004)
        
        Ip_av = Ip + Ip_av;
        Qp_av = Qp + Qp_av;
        
        # k = 0;
        t_seg = np.arange(0,len(Ip_av))*dt
        
        
        for pulseindx in range(len(pulseIRaws)):
            
            Iv = pulseIRaws[pulseindx]
            Qv = pulseQRaws[pulseindx]
            
            
            points = int((1/pulselaser/2 - t_offset)/dt)
            
            if calibrate_offset:
                Iv = Iv - np.mean(Iv[0:points]) + Ip_initial
                Qv = Qv - np.mean(Qv[0:points]) + Qp_initial
            
            t_bin,Iv_bined = ut.AverageArray(t_seg,Iv,chuncksize = binsize)
            t_bin,Qv_bined = ut.AverageArray(t_seg,Qv,chuncksize = binsize)
            pulse_amp,pulse_phase = res.cal_pulse_outside(Iv_bined + 1j*Qv_bined,pulsedata['freq']*1e9,pulsewidth = 50,fs = 1/(t_bin[1]-t_bin[0]))
            
            avenum = int((1/pulselaser/2 - t_offset)/binsize/dt)
            
            if pulsetype == 'phase':
                pulse_phase = np.unwrap(pulse_phase)
            
            
                phase0 = np.mean(pulse_phase[0:avenum])
            
                pulse_degree = (pulse_phase-phase0)
            
                pulse_degree = np.unwrap(pulse_degree)*180/np.pi
                pulse_wiener = sg.lfilter(wienerfilter,1,pulse_degree)
            
                indxpulse = len(wienerfilter)
            
                p_min = np.min(pulse_wiener[indxpulse-5:indxpulse+5])
            
                pulse_height.append(p_min)
            
                pulse_height_2.append(np.min(pulse_wiener[indxpulse-2:indxpulse+2]))
            
            else:
                
                amp0 = np.mean(pulse_amp[0:avenum])
            
                pulse_amp = (pulse_amp-amp0)
            
                pulse_amp = pulse_amp*180/np.pi
                pulse_wiener = sg.lfilter(wienerfilter,1,pulse_amp)
            
                indxpulse = len(wienerfilter)
            
                p_min = np.min(pulse_wiener[indxpulse-5:indxpulse+5])
            
                pulse_height.append(p_min)
            
                pulse_height_2.append(np.min(pulse_wiener[indxpulse-2:indxpulse+2]))
            

    # if pulsetype == 'phase':
    p_array = -np.array(pulse_height)
    pulsemin = np.min(p_array)
    pulsemax = np.max(p_array)
    # else:
        # p_array = -np.array(pulse_height_amp)
        # pulsemin = np.min(p_array)
        # pulsemax = np.max(p_array)
    
    bins = np.arange(pulsemin,pulsemax,0.2)
    
    plt.figure()
    plt.hist(p_array,bins = bins)    

    plt.xlabel('Pulse height (deg)',fontsize = 20)
    plt.ylabel('Counts',fontsize = 20)
    
    if pulsetype == 'phase':
        plt.savefig(savefolder + '/pulse statistics-average%d.eps'%(binsize),bbox_inches = 'tight')
        plt.savefig(savefolder + '/pulse statistics-average%d.jpg'%(binsize),bbox_inches = 'tight')
        plt.savefig(savefolder + '/pulse statistics-average%d.svg'%(binsize),bbox_inches = 'tight')
    
    else:
        plt.savefig(savefolder + '/pulse statistics-average-amp%d.eps'%(binsize),bbox_inches = 'tight')
        plt.savefig(savefolder + '/pulse statistics-average-amp%d.jpg'%(binsize),bbox_inches = 'tight')
        plt.savefig(savefolder + '/pulse statistics-average-amp%d.svg'%(binsize),bbox_inches = 'tight')
    
    
    # plt.close()
    
    if pulsetype == 'phase':
        np.savetxt(savefolder + '/pulse_height-average%d.txt'%(binsize),pulse_height)
        np.savetxt(savefolder + '/pulse_height-average%d_2.txt'%(binsize),pulse_height_2)
        
    else:
        np.savetxt(savefolder + '/pulse_height-average-amp%d.txt'%(binsize),pulse_height)
        np.savetxt(savefolder + '/pulse_height-average-amp%d_2.txt'%(binsize),pulse_height_2)
        
    
    return pulse_height

def Gaussian(x,a,mu,sigma):
    
    return a/sigma/np.sqrt(2*np.pi)*np.exp(-(x-mu)**2/2/sigma**2)



def EnergyResolution(pulseheight,savefolder,photonum = 4,binsize = 0.1,
                     peakwidth = 4,plotfigs = True,Format='.jpg'):
    
    pulseheight = np.array(pulseheight)
    
    
    pmin = np.min(-pulseheight)
    pmax = np.max(-pulseheight)

    # binsize = 0.1

    bins = np.arange(pmin,pmax+binsize ,binsize)

    a = np.histogram(-pulseheight,bins)

    x = a[1]
    x = x[:-1]
    y = a[0]

    vals = sg.find_peaks(y,width = peakwidth)

    peaks_indx = vals[0]

    center0 = x[peaks_indx[0]]
    dpeak = x[peaks_indx[1]] - center0


    gaussmodes = []
    for peak in range(photonum):
        
        prefix = 'g%d_'%(peak)
        gauss = GaussianModel(prefix = 'g%d_'%(peak))
        
        if peak == 0:
            pars = gauss.make_params()
        else:
            pars.update(gauss.make_params())
        
        if peak == 0:
            pmin = np.min(y)
            pars[prefix + 'center'].set(center0,min = pmin,max = center0 + dpeak/2)
            
            
        else:
            pars[prefix + 'center'].set(center0,min = center0 - dpeak/2,max = center0 + dpeak/2)
        
        center0 = center0 + dpeak
        
        pars[prefix + 'sigma'].set(1,min = 0.1,max = 2)
        pars[prefix + 'amplitude'].set(100,min = 0)
            
        gaussmodes.append(gauss)
        
        if peak == 0:
            mod = gauss;
        
        else:
            mod = mod + gauss
        

    out = mod.fit(y, pars, x=x)

    # print(out.fit_report(min_correl=0.5))
    
    if plotfigs == True:
        plt.figure()
    
        plt.hist(-pulseheight,bins)
        plt.plot(x + binsize/2, out.best_fit, 'r-',linewidth = 2)
    # plt.show()
        plt.xlabel('Pulse Height(deg)')
        plt.ylabel('Counts')

        deltaEs = []
        Es = []
        sigmas = []
        for i in range(photonum):
            amp = out.best_values['g%d_amplitude'%(i)]
            center = out.best_values['g%d_center'%(i)]
            sigma = out.best_values['g%d_sigma'%(i)]
        
            y = Gaussian(x,amp,center,sigma)
        
            deltaEs.append(2.3548200*sigma)
            Es.append(center)
            sigmas.append(sigma)
        
            plt.plot(x+binsize/2,y,'--',linewidth = 3,color = 'red',alpha = 0.5)
        
            peak = amp/sigma/np.sqrt(2*np.pi)
        
            plt.text(center + 0.5,peak+2,'n=%d'%(i),fontsize = 18)
        
        
        g0_center = out.best_values['g0_center']
        g1_center = out.best_values['g1_center']
        
        sigma1 = out.best_values['g1_sigma']
        
        R = 1/2/np.sqrt(2*np.log(2)) * (g1_center-g0_center)/sigma1
        # R = 1/2/np.sqrt(2*np.log(2)) * (g1_center)/sigma1
        
        plt.annotate('$\\frac{E}{\Delta E}$' + '= %1.1f @ n=1' %(R), xy=(1,1), xytext=(-130, -12), va='top',
                 xycoords='axes fraction', textcoords='offset points',fontsize = 20)

        plt.savefig(savefolder + '//Energy Resolution'+Format,bbox_inches = 'tight')
        # plt.savefig(savefolder + '//Energy Resolution.svg',bbox_inches = 'tight')
    
        # plt.savefig(savefolder + '//Energy Resolution.jpg',dpi = 600,bbox_inches = 'tight')
        # print('fig saved')
    
    return out,Es,sigmas
    
    
def GenSinglePhotonPulse(res,IQCaldata,Ers,Sigmas,wienerfilter,savefolder,
                         pulsefolder,pulseheight = None,binsize = 500,
                         calibrate_offset = True,dt = 1e-8,pulselaser = 250,
                         NumpulseUnscaled=80,t_offset = 0.001,genrawdataplot = True,
                         rawdataplotnum = 2,plot_figs = True):

    pulsePointNum = round(1/pulselaser/dt);
    Ip_av = np.zeros(pulsePointNum);
    Qp_av = np.zeros(pulsePointNum)

    # pulse_height = []
    # pulse_height_2 = []

    pulseprocess = NumpulseUnscaled;
    
    photonums = len(Ers) + 1
    
    Isp = [Ip_av]*photonums
    Qsp = [Qp_av]*photonums
    
    pulsenums = [0]*photonums
    
    if genrawdataplot == True:
        rawdataplotted = [0]*photonums
        # rawdataplotnum
    # sp_pulse = 
    pulseindx = 0;

    for i in range(pulseprocess):
        
        print("i:",i)

        f = open(pulsefolder + '/%d.obj'%(i),'rb')
        pulsedata = pickle.load(f)
        f.close()
        
        
        I0 = res.I0
        Q0 = res.Q0
        freq = res.freq
        
        
        I,Q,pulsefreq,fsample = ut.CalPulseRaw(pulsedata,IQCaldata,freq,I0,Q0,ispulse = True)
        # t = np.arange(0,len(I))*1e-08

        Ip,Qp,pulseIRaws,pulseQRaws = ut.GenAveragePulse(I,Q,dt = dt,
                                                         pulsefreq = pulselaser,t_offset = t_offset)
        
        resfreqindx = np.argmin(np.abs(res.freq-pulsefreq*1e9))
        
        
        if i==0:

            Ip_initial = res.I[resfreqindx]
            Qp_initial = res.Q[resfreqindx]
        
        #take the data of 2ms before the pulse for the noise calculation 
        # Ip1,Qp1,noiseIRaws,noiseQRaws = ut.GenAveragePulse(I,Q,dt = 1e-8,pulsefreq = pulselaser,
                                                           # t_offset = 0.000,segtime = 0.004)
        
        # Ip_av = Ip + Ip_av;
        # Qp_av = Qp + Qp_av;
        
        # k = 0;
        t_seg = np.arange(0,len(Ip_av))*dt
        for pulseindx in range(len(pulseIRaws)):
            
            Iv = pulseIRaws[pulseindx]
            Qv = pulseQRaws[pulseindx]
            
            if calibrate_offset:
                
                points = int((1/pulselaser/2 - t_offset)/dt)
                Iv = Iv - np.mean(Iv[0:points]) + Ip_initial
                Qv = Qv - np.mean(Qv[0:points]) + Qp_initial
            
            
            if pulseheight is None or genrawdataplot == True:
                
                t_bin,Iv_bined = ut.AverageArray(t_seg,Iv,chuncksize = binsize)
                t_bin,Qv_bined = ut.AverageArray(t_seg,Qv,chuncksize = binsize)
                pulse_amp,pulse_phase = res.cal_pulse_outside(Iv_bined + 1j*Qv_bined,pulsedata['freq']*1e9,pulsewidth = 50,fs = 1/(t_bin[1]-t_bin[0]))
                
                pulse_phase = np.unwrap(pulse_phase)
            
                avenum = int((1/pulselaser/2 - t_offset)/binsize/dt)
                phase0 = np.mean(pulse_phase[0:avenum])
            
                pulse_degree = (pulse_phase-phase0)
            
                pulse_degree = np.unwrap(pulse_degree)*180/np.pi
                pulse_wiener = sg.lfilter(wienerfilter,1,pulse_degree)
            
                indxpulse = len(wienerfilter)
            
                phCurrent = -np.min(pulse_wiener[indxpulse-5:indxpulse+5])
            
            else:
                
                phCurrent = -pulseheight[pulseindx]
                pulseindx += 1
            
            photonum = -1
            for indx,(er,sigma) in enumerate(zip(Ers,Sigmas)):
                
                if indx == 0:
                    if phCurrent < er + sigma*2.355/2:
                        
                        Isp[indx] = Isp[indx] + Iv
                        Qsp[indx] = Qsp[indx] + Qv
                        
                        pulsenums[indx] +=  1
                        photonum = 0
                        
                        break
                        
                elif indx <= len(Ers) - 1:
                    if phCurrent > er - sigma*2.355/2 and phCurrent < er + sigma*2.355/2:
                        
                        Isp[indx] = Isp[indx] + Iv
                        Qsp[indx] = Qsp[indx] + Qv
                        
                        pulsenums[indx] +=  1
                        photonum = indx
                        
                        break
                    
                else:
                    if phCurrent > Ers[-1] + Sigmas[-1]*2.355/2:
                        Isp[indx] = Isp[indx] + Iv
                        Qsp[indx] = Qsp[indx] + Qv
                    
                        pulsenums[indx] +=  1
                        photonum = indx
                    break
            
            if photonum == -1:
                continue
            
            if genrawdataplot == True:
                
               
                if rawdataplotted[photonum] == rawdataplotnum:
                    # rawdataplotted[photonum] = 
                    continue
                else:
                    rawdataplotted[photonum] += 1
                
                
                plt.figure()
                plt.plot(t_bin*1000,pulse_degree,color = 'blue',label = 'raw')
                plt.plot(t_bin*1000,pulse_wiener,color = 'red',label = 'wiener filtered')
                
                plt.xlabel('Time(ms)')
                plt.ylabel('Pulse Height(deg)')
                
                plt.legend(fontsize = 15)
                plt.title("%d Photon Event" %(photonum),fontsize = 20)
                
                plt.savefig(savefolder + '//%d Photon Event_%d.eps' %(photonum,rawdataplotted[photonum]),bbox_inches = 'tight')
                plt.savefig(savefolder + '//%d Photon Event_%d.jpg' %(photonum,rawdataplotted[photonum]),dpi = 600,bbox_inches = 'tight')
                
                # plt.close()
        
        #average the pulse number
        
    for i in reversed(range(len(pulsenums))):
        
        if pulsenums[i] == 0:
            continue
            
        Isp[i] = Isp[i]/pulsenums[i]
        Qsp[i] = Qsp[i]/pulsenums[i]
        
        
    res.noiseUnscaled = []
        
    data = {'Isp':Isp,
            'Qsp':Qsp,
            'Res':res,
            'pulsefreq':pulsedata['freq'],
            't_offset':t_offset,
            'pulsetriggertime':1/pulselaser/2 - t_offset,
            'dt':dt}
        
    f = open(savefolder + '/singlephotondata.obj','wb')
        
    pickle.dump(data,f)
        
    f.close()
    
    if plot_figs == True:
        
        PlotSingplePhotonPulse(data,savefolder,closefig = True)
        # plt_colors, sm = ut.GenColorMap(range(len(pulsenums)))
        # plt.figure()
                    
        # for i in reversed(range(len(pulsenums))):
        
        #     if pulsenums[i] == 0:
        #         continue
        #     pulse_amp,pulse_phase = res.cal_pulse_outside(Isp[i] + 1j*Qsp[i],pulsedata['freq']*1e9,pulsewidth = 50,fs = 1/(t_bin[1]-t_bin[0]))

        #     avenum = int((1/pulselaser/2 - t_offset)/dt)
        #     pulse_phase = np.unwrap(pulse_phase)
        #     phase0 = np.mean(pulse_phase[0:avenum])
        
        #     pulse_phase = np.unwrap(pulse_phase - phase0)
        
        #     plt.plot(t_seg*1000-t_offset*1000,(pulse_phase)*180/np.pi,
        #          label = '%d photons'%(i),color = plt_colors[i])
            
        #     plt.xlabel('Time(ms)')
        #     plt.ylabel('Pulse Height(deg)')
        #     plt.legend(fontsize = 15)
        
        
        # plt.savefig(savefolder + '//Single Photon Average.eps',bbox_inches = 'tight')
        # plt.savefig(savefolder + '//Single Photon Average.jpg',dpi = 600,bbox_inches = 'tight')
            
        # plt.close()
        
def PlotSingplePhotonPulse(spdata,savefolder,closefig = True):
    
    # data = {'Isp':Isp,
    #         'Qsp':Qsp,
    #         'Res':res,
    #         'pulsefreq':pulsedata['freq'],
    #         't_offset':t_offset,
    #         'pulsetriggertime':1/pulselaser/2 - t_offset,
    #         'dt':dt}
        
    Isp = spdata['Isp']
    Qsp = spdata['Qsp']
    res = spdata['Res']
    t_trigger =  spdata['pulsetriggertime']
    t_offset = spdata['t_offset']
    dt = spdata['dt']
    pulsefreq = spdata['pulsefreq']
    
    avenum = int(t_trigger/dt)
    
    plt_colors, sm = ut.GenColorMap(range(len(Isp)-1))
    fig, ax1 = plt.subplots()
    t_seg = np.arange(0,len(Isp[0]))*dt
    
    pmax = []
    for i in reversed(range(len(Isp))):
    
        # if pulsenums[i] == 0:
            # continue
        
        if np.sum(np.abs(Isp[i])) < 1e-10:
            continue
        
        pulse_amp,pulse_phase = res.cal_pulse_outside(Isp[i] + 1j*Qsp[i],pulsefreq*1e9,pulsewidth = 50,fs = 1/dt)

        # avenum = int((1/pulselaser/2 - t_offset)/dt)
        pulse_phase = np.unwrap(pulse_phase)
        
        phase0 = np.mean(pulse_phase[0:avenum])
    
        pulse_phase = np.unwrap(pulse_phase - phase0)
        
        pmax.append(np.min(pulse_phase))
    
        ax1.plot(t_seg*1000-t_offset*1000,(pulse_phase)*180/np.pi,
             label = '%d photons'%(i),color = plt_colors[i])
        
    ax1.set_xlabel('Time(ms)')
    ax1.set_ylabel('Pulse Height(deg)')
        # ax1.legend(fontsize = 15)
    cbar = plt.colorbar(sm,ax = ax1)
    cbar.ax.set_ylabel('Photon number', rotation=270,labelpad = 20)
    photonum = np.array([1,2,3,4])
    
    pmax.reverse()
    pmax = np.array(pmax)
    
    a = np.polyfit(photonum,pmax[1:-1]*180/np.pi,1)

    photonum = np.array([0,1,2,3,4,5])

    # plt.plot(photonum,a[0]*photonum + a[1])
    
    left, bottom, width,height = [0.42,0.3,0.3,0.3]
    ax2 = fig.add_axes([left,bottom,width,height])
    
    ax2.plot(pmax*180/np.pi,'s')
    ax2.plot(photonum,(a[0]*photonum + a[1]))
    ax2.set_xlabel('Photo number',fontsize = 12)
    ax2.set_ylabel('Pulse Height(deg)',fontsize = 12)
    
    plt.setp(ax2.get_xticklabels(), fontsize=12)
    plt.setp(ax2.get_yticklabels(), fontsize=12)
    
    plt.savefig(savefolder + '//Single Photon Average.eps',bbox_inches = 'tight')
    plt.savefig(savefolder + '//Single Photon Average.jpg',dpi = 600,bbox_inches = 'tight')
    
    # if closefig == True:
    #     plt.close()     
        
    
        
    return fig,pmax

